import logging
import string

logger = logging.getLogger(__name__)

try:
    import dbus
except ImportError:
    dbus = None

_AVAHI_IF_UNSPEC = -1
_AVAHI_PROTO_UNSPEC = -1
_AVAHI_PUBLISHFLAGS_NONE = 0


def _is_loopback_address(host):
    return (
        host.startswith("127.")
        or host.startswith("::ffff:127.")
        or host == "::1"
    )


def _convert_text_list_to_dbus_format(text_list):
    array = dbus.Array(signature="ay")
    for text in text_list:
        array.append([dbus.Byte(ord(c)) for c in text])
    return array


class Zeroconf:

    """Publish a network service with Zeroconf.

    Currently, this only works on Linux using Avahi via D-Bus.

    :param str name: human readable name of the service, e.g. 'MPD on neptune'
    :param str stype: service type, e.g. '_mpd._tcp'
    :param int port: TCP port of the service, e.g. 6600
    :param str domain: local network domain name, defaults to ''
    :param str host: interface to advertise the service on, defaults to ''
    :param text: extra information depending on ``stype``, defaults to empty
        list
    :type text: list of str
    """

    def __init__(self, name, stype, port, domain="", host="", text=None):
        self.stype = stype
        self.port = port
        self.domain = domain
        self.host = host
        self.text = text or []

        self.bus = None
        self.server = None
        self.group = None
        self.display_hostname = None
        self.name = None

        if dbus:
            try:
                self.bus = dbus.SystemBus()
                self.server = dbus.Interface(
                    self.bus.get_object("org.freedesktop.Avahi", "/"),
                    "org.freedesktop.Avahi.Server",
                )
                self.display_hostname = f"{self.server.GetHostName()}"
                self.name = string.Template(name).safe_substitute(
                    hostname=self.display_hostname, port=port
                )
            except dbus.exceptions.DBusException as e:
                logger.debug("%s: Server failed: %s", self, e)

    def __str__(self):
        return (
            f"Zeroconf service {self.name!r} "
            f"({self.stype} at [{self.host}]:{self.port:d})"
        )

    def publish(self):
        """Publish the service.

        Call when your service starts.
        """

        if _is_loopback_address(self.host):
            logger.debug(
                "%s: Publish on loopback interface is not supported.", self
            )
            return False

        if not dbus:
            logger.debug("%s: dbus not installed; publish failed.", self)
            return False

        if not self.bus:
            logger.debug("%s: Bus not available; publish failed.", self)
            return False

        if not self.server:
            logger.debug("%s: Server not available; publish failed.", self)
            return False

        try:
            if not self.bus.name_has_owner("org.freedesktop.Avahi"):
                logger.debug(
                    "%s: Avahi service not running; publish failed.", self
                )
                return False

            self.group = dbus.Interface(
                self.bus.get_object(
                    "org.freedesktop.Avahi", self.server.EntryGroupNew()
                ),
                "org.freedesktop.Avahi.EntryGroup",
            )

            self.group.AddService(
                _AVAHI_IF_UNSPEC,
                _AVAHI_PROTO_UNSPEC,
                dbus.UInt32(_AVAHI_PUBLISHFLAGS_NONE),
                self.name,
                self.stype,
                self.domain,
                self.host,
                dbus.UInt16(self.port),
                _convert_text_list_to_dbus_format(self.text),
            )

            self.group.Commit()
            logger.debug("%s: Published", self)
            return True
        except dbus.exceptions.DBusException as e:
            logger.debug("%s: Publish failed: %s", self, e)
            return False

    def unpublish(self):
        """Unpublish the service.

        Call when your service shuts down.
        """

        if self.group:
            try:
                self.group.Reset()
                logger.debug("%s: Unpublished", self)
            except dbus.exceptions.DBusException as e:
                logger.debug("%s: Unpublish failed: %s", self, e)
            finally:
                self.group = None



import unittest

from mopidy.core import HistoryController
from mopidy.internal.models import HistoryState, HistoryTrack
from mopidy.models import Artist, Ref, Track


class PlaybackHistoryTest(unittest.TestCase):
    def setUp(self):  # noqa: N802
        self.tracks = [
            Track(
                uri="dummy1:a",
                name="foo",
                artists=[Artist(name="foober"), Artist(name="barber")],
            ),
            Track(uri="dummy2:a", name="foo"),
            Track(uri="dummy3:a", name="bar"),
        ]
        self.history = HistoryController()

    def test_add_track(self):
        self.history._add_track(self.tracks[0])
        assert self.history.get_length() == 1

        self.history._add_track(self.tracks[1])
        assert self.history.get_length() == 2

        self.history._add_track(self.tracks[2])
        assert self.history.get_length() == 3

    def test_non_tracks_are_rejected(self):
        with self.assertRaises(TypeError):
            self.history._add_track(object())

        assert self.history.get_length() == 0

    def test_history_entry_contents(self):
        track = self.tracks[0]
        self.history._add_track(track)

        result = self.history.get_history()
        (timestamp, ref) = result[0]

        assert isinstance(timestamp, int)
        assert track.uri == ref.uri
        assert track.name in ref.name
        for artist in track.artists:
            assert artist.name in ref.name


class CoreHistorySaveLoadStateTest(unittest.TestCase):
    def setUp(self):  # noqa: N802
        self.tracks = [
            Track(uri="dummy1:a", name="foober"),
            Track(uri="dummy2:a", name="foo"),
            Track(uri="dummy3:a", name="bar"),
        ]

        self.refs = []
        for t in self.tracks:
            self.refs.append(Ref.track(uri=t.uri, name=t.name))

        self.history = HistoryController()

    def test_save(self):
        self.history._add_track(self.tracks[2])
        self.history._add_track(self.tracks[1])

        value = self.history._save_state()

        assert len(value.history) == 2
        # last in, first out
        assert value.history[0].track == self.refs[1]
        assert value.history[1].track == self.refs[2]

    def test_load(self):
        state = HistoryState(
            history=[
                HistoryTrack(timestamp=34, track=self.refs[0]),
                HistoryTrack(timestamp=45, track=self.refs[2]),
                HistoryTrack(timestamp=56, track=self.refs[1]),
            ]
        )
        coverage = ["history"]
        self.history._load_state(state, coverage)

        hist = self.history.get_history()
        assert len(hist) == 3
        assert hist[0] == (34, self.refs[0])
        assert hist[1] == (45, self.refs[2])
        assert hist[2] == (56, self.refs[1])

        # after import, adding more tracks must be possible
        self.history._add_track(self.tracks[1])
        hist = self.history.get_history()
        assert len(hist) == 4
        assert hist[0][1] == self.refs[1]
        assert hist[1] == (34, self.refs[0])
        assert hist[2] == (45, self.refs[2])
        assert hist[3] == (56, self.refs[1])

    def test_load_invalid_type(self):
        with self.assertRaises(TypeError):
            self.history._load_state(11, None)

    def test_load_none(self):
        self.history._load_state(None, None)



# TODO Test browse()
from unittest import mock

import pytest

from mopidy.file import backend
from mopidy.internal import path

from tests import path_to_data_dir


@pytest.fixture
def config():
    return {
        "proxy": {},
        "file": {
            "show_dotfiles": False,
            "media_dirs": [str(path_to_data_dir(""))],
            "excluded_file_extensions": [],
            "follow_symlinks": False,
            "metadata_timeout": 1000,
        },
    }


@pytest.fixture
def audio():
    return mock.Mock()


@pytest.fixture
def track_uri():
    return path.path_to_uri(path_to_data_dir("song1.wav"))


def test_file_browse(config, audio, track_uri, caplog):
    provider = backend.FileBackend(audio=audio, config=config).library
    result = provider.browse(track_uri)

    assert len(result) == 0
    assert any(map(lambda record: record.levelname == "ERROR", caplog.records))
    assert any(
        map(lambda record: "Rejected attempt" in record.message, caplog.records)
    )



import argparse
import collections
import contextlib
import logging
import os
import pathlib
import signal
import sys

import pykka
from pykka.messages import ProxyCall

from mopidy import config as config_lib
from mopidy import exceptions
from mopidy.audio import Audio
from mopidy.core import Core
from mopidy.internal import deps, process, timer, versioning
from mopidy.internal.gi import GLib

logger = logging.getLogger(__name__)

_default_config = [
    (pathlib.Path(base) / "mopidy" / "mopidy.conf").resolve()
    for base in GLib.get_system_config_dirs() + [GLib.get_user_config_dir()]
]
DEFAULT_CONFIG = ":".join(map(str, _default_config))


def config_files_type(value):
    return value.split(":")


def config_override_type(value):
    try:
        section, remainder = value.split("/", 1)
        key, value = remainder.split("=", 1)
        return (section.strip(), key.strip(), value.strip())
    except ValueError:
        raise argparse.ArgumentTypeError(
            f"{value} must have the format section/key=value"
        )


class _ParserError(Exception):
    def __init__(self, message):
        self.message = message


class _HelpError(Exception):
    pass


class _ArgumentParser(argparse.ArgumentParser):
    def error(self, message):
        raise _ParserError(message)


class _HelpAction(argparse.Action):
    def __init__(self, option_strings, dest=None, help=None):
        super().__init__(
            option_strings=option_strings,
            dest=dest or argparse.SUPPRESS,
            default=argparse.SUPPRESS,
            nargs=0,
            help=help,
        )

    def __call__(self, parser, namespace, values, option_string=None):
        raise _HelpError()


class Command:

    """Command parser and runner for building trees of commands.

    This class provides a wraper around :class:`argparse.ArgumentParser`
    for handling this type of command line application in a better way than
    argprases own sub-parser handling.
    """

    help = None
    #: Help text to display in help output.

    def __init__(self):
        self._children = collections.OrderedDict()
        self._arguments = []
        self._overrides = {}

    def _build(self):
        actions = []
        parser = _ArgumentParser(add_help=False)
        parser.register("action", "help", _HelpAction)

        for args, kwargs in self._arguments:
            actions.append(parser.add_argument(*args, **kwargs))

        parser.add_argument(
            "_args", nargs=argparse.REMAINDER, help=argparse.SUPPRESS
        )
        return parser, actions

    def add_child(self, name, command):
        """Add a child parser to consider using.

        :param name: name to use for the sub-command that is being added.
        :type name: string
        """
        self._children[name] = command

    def add_argument(self, *args, **kwargs):
        """Add an argument to the parser.

        This method takes all the same arguments as the
        :class:`argparse.ArgumentParser` version of this method.
        """
        self._arguments.append((args, kwargs))

    def set(self, **kwargs):
        """Override a value in the finaly result of parsing."""
        self._overrides.update(kwargs)

    def exit(self, status_code=0, message=None, usage=None):
        """Optionally print a message and exit."""
        print("\n\n".join(m for m in (usage, message) if m))
        sys.exit(status_code)

    def format_usage(self, prog=None):
        """Format usage for current parser."""
        actions = self._build()[1]
        prog = prog or os.path.basename(sys.argv[0])
        return self._usage(actions, prog) + "\n"

    def _usage(self, actions, prog):
        formatter = argparse.HelpFormatter(prog)
        formatter.add_usage(None, actions, [])
        return formatter.format_help().strip()

    def format_help(self, prog=None):
        """Format help for current parser and children."""
        actions = self._build()[1]
        prog = prog or os.path.basename(sys.argv[0])

        formatter = argparse.HelpFormatter(prog)
        formatter.add_usage(None, actions, [])

        if self.help:
            formatter.add_text(self.help)

        if actions:
            formatter.add_text("OPTIONS:")
            formatter.start_section(None)
            formatter.add_arguments(actions)
            formatter.end_section()

        subhelp = []
        for name, child in self._children.items():
            child._subhelp(name, subhelp)

        if subhelp:
            formatter.add_text("COMMANDS:")
            subhelp.insert(0, "")

        return formatter.format_help() + "\n".join(subhelp)

    def _subhelp(self, name, result):
        actions = self._build()[1]

        if self.help or actions:
            formatter = argparse.HelpFormatter(name)
            formatter.add_usage(None, actions, [], "")
            formatter.start_section(None)
            formatter.add_text(self.help)
            formatter.start_section(None)
            formatter.add_arguments(actions)
            formatter.end_section()
            formatter.end_section()
            result.append(formatter.format_help())

        for childname, child in self._children.items():
            child._subhelp(" ".join((name, childname)), result)

    def parse(self, args, prog=None):
        """Parse command line arguments.

        Will recursively parse commands until a final parser is found or an
        error occurs. In the case of errors we will print a message and exit.
        Otherwise, any overrides are applied and the current parser stored
        in the command attribute of the return value.

        :param args: list of arguments to parse
        :type args: list of strings
        :param prog: name to use for program
        :type prog: string
        :rtype: :class:`argparse.Namespace`
        """
        prog = prog or os.path.basename(sys.argv[0])
        try:
            return self._parse(
                args, argparse.Namespace(), self._overrides.copy(), prog
            )
        except _HelpError:
            self.exit(0, self.format_help(prog))

    def _parse(self, args, namespace, overrides, prog):
        overrides.update(self._overrides)
        parser, actions = self._build()

        try:
            result = parser.parse_args(args, namespace)
        except _ParserError as exc:
            self.exit(1, str(exc), self._usage(actions, prog))

        if not result._args:
            for attr, value in overrides.items():
                setattr(result, attr, value)
            delattr(result, "_args")
            result.command = self
            return result

        child = result._args.pop(0)
        if child not in self._children:
            usage = self._usage(actions, prog)
            self.exit(1, f"unrecognized command: {child}", usage)

        return self._children[child]._parse(
            result._args, result, overrides, " ".join([prog, child])
        )

    def run(self, *args, **kwargs):
        """Run the command.

        Must be implemented by sub-classes that are not simply an intermediate
        in the command namespace.
        """
        raise NotImplementedError


@contextlib.contextmanager
def _actor_error_handling(name):
    try:
        yield
    except exceptions.BackendError as exc:
        logger.error("Backend (%s) initialization error: %s", name, exc)
    except exceptions.FrontendError as exc:
        logger.error("Frontend (%s) initialization error: %s", name, exc)
    except exceptions.MixerError as exc:
        logger.error("Mixer (%s) initialization error: %s", name, exc)
    except Exception:
        logger.exception("Got un-handled exception from %s", name)


# TODO: move out of this utility class
class RootCommand(Command):
    def __init__(self):
        super().__init__()
        self.set(base_verbosity_level=0)
        self.add_argument(
            "-h", "--help", action="help", help="Show this message and exit"
        )
        self.add_argument(
            "--version",
            action="version",
            version=f"Mopidy {versioning.get_version()}",
        )
        self.add_argument(
            "-q",
            "--quiet",
            action="store_const",
            const=-1,
            dest="verbosity_level",
            help="less output (warning level)",
        )
        self.add_argument(
            "-v",
            "--verbose",
            action="count",
            dest="verbosity_level",
            default=0,
            help="more output (repeat up to 4 times for even more)",
        )
        self.add_argument(
            "--config",
            action="store",
            dest="config_files",
            type=config_files_type,
            default=DEFAULT_CONFIG,
            metavar="FILES",
            help="config files to use, colon seperated, later files override",
        )
        self.add_argument(
            "-o",
            "--option",
            action="append",
            dest="config_overrides",
            type=config_override_type,
            metavar="OPTIONS",
            help="`section/key=value` values to override config options",
        )

    def run(self, args, config):
        def on_sigterm(loop):
            logger.info("GLib mainloop got SIGTERM. Exiting...")
            loop.quit()

        loop = GLib.MainLoop()
        GLib.unix_signal_add(
            GLib.PRIORITY_DEFAULT, signal.SIGTERM, on_sigterm, loop
        )

        mixer_class = self.get_mixer_class(config, args.registry["mixer"])
        backend_classes = args.registry["backend"]
        frontend_classes = args.registry["frontend"]
        core = None

        exit_status_code = 0
        try:
            mixer = None
            if mixer_class is not None:
                mixer = self.start_mixer(config, mixer_class)
            if mixer:
                self.configure_mixer(config, mixer)
            audio = self.start_audio(config, mixer)
            backends = self.start_backends(config, backend_classes, audio)
            core = self.start_core(config, mixer, backends, audio)
            self.start_frontends(config, frontend_classes, core)
            logger.info("Starting GLib mainloop")
            loop.run()
        except (
            exceptions.BackendError,
            exceptions.FrontendError,
            exceptions.MixerError,
        ):
            logger.info("Initialization error. Exiting...")
            exit_status_code = 1
        except KeyboardInterrupt:
            logger.info("Interrupted. Exiting...")
        except Exception:
            logger.exception("Uncaught exception")
        finally:
            loop.quit()
            self.stop_frontends(frontend_classes)
            self.stop_core(core)
            self.stop_backends(backend_classes)
            self.stop_audio()
            if mixer_class is not None:
                self.stop_mixer(mixer_class)
            process.stop_remaining_actors()
        return exit_status_code

    def get_mixer_class(self, config, mixer_classes):
        logger.debug(
            "Available Mopidy mixers: %s",
            ", ".join(m.__name__ for m in mixer_classes) or "none",
        )

        if config["audio"]["mixer"] == "none":
            logger.debug("Mixer disabled")
            return None

        selected_mixers = [
            m for m in mixer_classes if m.name == config["audio"]["mixer"]
        ]
        if len(selected_mixers) != 1:
            logger.error(
                'Did not find unique mixer "%s". Alternatives are: %s',
                config["audio"]["mixer"],
                ", ".join(m.name for m in mixer_classes) + ", none" or "none",
            )
            process.exit_process()
        return selected_mixers[0]

    def start_mixer(self, config, mixer_class):
        logger.info("Starting Mopidy mixer: %s", mixer_class.__name__)
        with _actor_error_handling(mixer_class.__name__):
            mixer = mixer_class.start(config=config).proxy()
            try:
                mixer.ping().get()
                return mixer
            except pykka.ActorDeadError as exc:
                logger.error("Actor died: %s", exc)
        return None

    def configure_mixer(self, config, mixer):
        volume = config["audio"]["mixer_volume"]
        if volume is not None:
            mixer.set_volume(volume)
            logger.info("Mixer volume set to %d", volume)
        else:
            logger.debug("Mixer volume left unchanged")

    def start_audio(self, config, mixer):
        logger.info("Starting Mopidy audio")
        return Audio.start(config=config, mixer=mixer).proxy()

    def start_backends(self, config, backend_classes, audio):
        logger.info(
            "Starting Mopidy backends: %s",
            ", ".join(b.__name__ for b in backend_classes) or "none",
        )

        backends = []
        for backend_class in backend_classes:
            with _actor_error_handling(backend_class.__name__):
                with timer.time_logger(backend_class.__name__):
                    backend = backend_class.start(
                        config=config, audio=audio
                    ).proxy()
                    backends.append(backend)

        # Block until all on_starts have finished, letting them run in parallel
        for backend in backends[:]:
            try:
                backend.ping().get()
            except pykka.ActorDeadError as exc:
                backends.remove(backend)
                logger.error("Actor died: %s", exc)

        return backends

    def start_core(self, config, mixer, backends, audio):
        logger.info("Starting Mopidy core")
        core = Core.start(
            config=config, mixer=mixer, backends=backends, audio=audio
        ).proxy()
        call = ProxyCall(attr_path=["_setup"], args=[], kwargs={})
        core.actor_ref.ask(call, block=True)
        return core

    def start_frontends(self, config, frontend_classes, core):
        logger.info(
            "Starting Mopidy frontends: %s",
            ", ".join(f.__name__ for f in frontend_classes) or "none",
        )

        for frontend_class in frontend_classes:
            with _actor_error_handling(frontend_class.__name__):
                with timer.time_logger(frontend_class.__name__):
                    frontend_class.start(config=config, core=core)

    def stop_frontends(self, frontend_classes):
        logger.info("Stopping Mopidy frontends")
        for frontend_class in frontend_classes:
            process.stop_actors_by_class(frontend_class)

    def stop_core(self, core):
        logger.info("Stopping Mopidy core")
        if core:
            call = ProxyCall(attr_path=["_teardown"], args=[], kwargs={})
            core.actor_ref.ask(call, block=True)
        process.stop_actors_by_class(Core)

    def stop_backends(self, backend_classes):
        logger.info("Stopping Mopidy backends")
        for backend_class in backend_classes:
            process.stop_actors_by_class(backend_class)

    def stop_audio(self):
        logger.info("Stopping Mopidy audio")
        process.stop_actors_by_class(Audio)

    def stop_mixer(self, mixer_class):
        logger.info("Stopping Mopidy mixer")
        process.stop_actors_by_class(mixer_class)


class ConfigCommand(Command):
    help = "Show currently active configuration."

    def __init__(self):
        super().__init__()
        self.set(base_verbosity_level=-1)

    def run(self, config, errors, schemas):
        data = config_lib.format(config, schemas, errors)

        # Throw away all bytes that are not valid UTF-8 before printing
        data = data.encode(errors="surrogateescape").decode(errors="replace")

        print(data)
        return 0


class DepsCommand(Command):
    help = "Show dependencies and debug information."

    def __init__(self):
        super().__init__()
        self.set(base_verbosity_level=-1)

    def run(self):
        print(deps.format_dependency_list())
        return 0



import pathlib


def path_to_data_dir(name):
    path = pathlib.Path(__file__).parent / "data" / name
    return path.resolve()


class IsA:
    def __init__(self, klass):
        self.klass = klass

    def __eq__(self, rhs):
        try:
            return isinstance(rhs, self.klass)
        except TypeError:
            return type(rhs) == type(self.klass)  # noqa

    def __ne__(self, rhs):
        return not self.__eq__(rhs)

    def __repr__(self):
        return str(self.klass)


any_int = IsA(int)
any_str = IsA(str)
any_unicode = any_str  # TODO remove



"""Mopidy documentation build configuration file"""


import os
import sys

sys.path.insert(0, os.path.abspath(os.path.dirname(__file__)))
sys.path.insert(0, os.path.abspath(os.path.dirname(__file__) + "/../"))

from mopidy.internal.versioning import get_version  # isort:skip  # noqa


# -- Workarounds to have autodoc generate API docs ----------------------------


class Mock:
    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        return Mock()

    def __or__(self, other):
        return Mock()

    def __mro_entries__(self, bases):
        return tuple()

    @classmethod
    def __getattr__(cls, name):
        if name == "get_system_config_dirs":  # GLib.get_system_config_dirs()
            return list
        elif name == "get_user_config_dir":  # GLib.get_user_config_dir()
            return str
        else:
            return Mock()


MOCK_MODULES = [
    "dbus",
    "dbus.mainloop",
    "dbus.mainloop.glib",
    "dbus.service",
    "mopidy.internal.gi",
]
for mod_name in MOCK_MODULES:
    sys.modules[mod_name] = Mock()


# -- Custom Sphinx object types -----------------------------------------------


def setup(app):
    from sphinx.ext.autodoc import cut_lines

    app.connect("autodoc-process-docstring", cut_lines(4, what=["module"]))
    app.add_object_type(
        "confval",
        "confval",
        objname="configuration value",
        indextemplate="pair: %s; configuration value",
    )


# -- General configuration ----------------------------------------------------

needs_sphinx = "1.3"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.extlinks",
    "sphinx.ext.graphviz",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
]

templates_path = ["_templates"]
source_suffix = ".rst"
master_doc = "index"

project = "Mopidy"
copyright = "2009-2020, Stein Magnus Jodal and contributors"


release = get_version()
version = ".".join(release.split(".")[:2])

# To make the build reproducible, avoid using today's date in the manpages
today = "2020"

exclude_trees = ["_build"]

pygments_style = "sphinx"

modindex_common_prefix = ["mopidy."]


# -- Options for HTML output --------------------------------------------------

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]

html_use_modindex = True
html_use_index = True
html_split_index = False
html_show_sourcelink = True

htmlhelp_basename = "Mopidy"


# -- Options for LaTeX output -------------------------------------------------

latex_documents = [
    (
        "index",
        "Mopidy.tex",
        "Mopidy Documentation",
        "Stein Magnus Jodal and contributors",
        "manual",
    ),
]


# -- Options for manpages output ----------------------------------------------

man_pages = [
    ("command", "mopidy", "music server", "", "1"),
]


# -- Options for extlink extension --------------------------------------------

extlinks = {
    "issue": ("https://github.com/mopidy/mopidy/issues/%s", "#"),
    "commit": ("https://github.com/mopidy/mopidy/commit/%s", "commit "),
    "js": ("https://github.com/mopidy/mopidy.js/issues/%s", "mopidy.js#"),
    "mpris": (
        "https://github.com/mopidy/mopidy-mpris/issues/%s",
        "mopidy-mpris#",
    ),
    "discuss": ("https://discourse.mopidy.com/t/%s", "discourse.mopidy.com/t/"),
}


# -- Options for intersphinx extension ----------------------------------------

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "pykka": ("https://www.pykka.org/en/latest/", None),
    "tornado": ("http://www.tornadoweb.org/en/stable/", None),
}

# -- Options for linkcheck builder -------------------------------------------

linkcheck_ignore = [  # Some sites work in browser but linkcheck fails.
    r"http://localhost:\d+/",
    r"http://wiki.commonjs.org",
    r"http://vk.com",
    r"http://$",
]

linkcheck_anchors = False  # This breaks on links that use # for other stuff



import logging
import os

from mopidy import backend, exceptions, models
from mopidy.audio import scan, tags
from mopidy.internal import path

logger = logging.getLogger(__name__)


class FileLibraryProvider(backend.LibraryProvider):
    """Library for browsing local files."""

    # TODO: get_images that can pull from metadata and/or .folder.png etc?
    # TODO: handle playlists?

    @property
    def root_directory(self):
        if not self._media_dirs:
            return None
        elif len(self._media_dirs) == 1:
            uri = path.path_to_uri(self._media_dirs[0]["path"])
        else:
            uri = "file:root"
        return models.Ref.directory(name="Files", uri=uri)

    def __init__(self, backend, config):
        super().__init__(backend)
        self._media_dirs = list(self._get_media_dirs(config))
        self._show_dotfiles = config["file"]["show_dotfiles"]
        self._excluded_file_extensions = tuple(
            file_ext.lower()
            for file_ext in config["file"]["excluded_file_extensions"]
        )
        self._follow_symlinks = config["file"]["follow_symlinks"]

        self._scanner = scan.Scanner(timeout=config["file"]["metadata_timeout"])

    def browse(self, uri):
        logger.debug("Browsing files at: %s", uri)
        result = []
        local_path = path.uri_to_path(uri)

        if str(local_path) == "root":
            return list(self._get_media_dirs_refs())

        if not self._is_in_basedir(local_path):
            logger.warning(
                "Rejected attempt to browse path (%s) outside dirs defined "
                "in file/media_dirs config.",
                uri,
            )
            return []
        if path.uri_to_path(uri).is_file():
            logger.error("Rejected attempt to browse file (%s)", uri)
            return []

        for dir_entry in local_path.iterdir():
            child_path = dir_entry.resolve()
            uri = path.path_to_uri(child_path)

            if not self._show_dotfiles and dir_entry.name.startswith("."):
                continue

            if (
                self._excluded_file_extensions
                and dir_entry.suffix in self._excluded_file_extensions
            ):
                continue

            if child_path.is_symlink() and not self._follow_symlinks:
                logger.debug("Ignoring symlink: %s", uri)
                continue

            if not self._is_in_basedir(child_path):
                logger.debug("Ignoring symlink to outside base dir: %s", uri)
                continue

            if child_path.is_dir():
                result.append(
                    models.Ref.directory(name=dir_entry.name, uri=uri)
                )
            elif child_path.is_file():
                result.append(models.Ref.track(name=dir_entry.name, uri=uri))

        def order(item):
            return (item.type != models.Ref.DIRECTORY, item.name)

        result.sort(key=order)

        return result

    def lookup(self, uri):
        logger.debug("Looking up file URI: %s", uri)
        local_path = path.uri_to_path(uri)

        try:
            result = self._scanner.scan(uri)
            track = tags.convert_tags_to_track(result.tags).replace(
                uri=uri, length=result.duration
            )
        except exceptions.ScannerError as e:
            logger.warning("Failed looking up %s: %s", uri, e)
            track = models.Track(uri=uri)

        if not track.name:
            track = track.replace(name=local_path.name)

        return [track]

    def _get_media_dirs(self, config):
        for entry in config["file"]["media_dirs"]:
            media_dir = {}
            media_dir_split = entry.split("|", 1)
            local_path = path.expand_path(media_dir_split[0])

            if local_path is None:
                logger.debug(
                    "Failed expanding path (%s) from file/media_dirs config "
                    "value.",
                    media_dir_split[0],
                )
                continue
            elif not local_path.is_dir():
                logger.warning(
                    "%s is not a directory. Please create the directory or "
                    "update the file/media_dirs config value.",
                    local_path,
                )
                continue

            media_dir["path"] = local_path
            if len(media_dir_split) == 2:
                media_dir["name"] = media_dir_split[1]
            else:
                # TODO Mpd client should accept / in dir name
                media_dir["name"] = media_dir_split[0].replace(os.sep, "+")

            yield media_dir

    def _get_media_dirs_refs(self):
        for media_dir in self._media_dirs:
            yield models.Ref.directory(
                name=media_dir["name"], uri=path.path_to_uri(media_dir["path"])
            )

    def _is_in_basedir(self, local_path):
        return any(
            path.is_path_inside_base_dir(local_path, media_dir["path"])
            for media_dir in self._media_dirs
        )



import unittest

from mopidy import exceptions


class ExceptionsTest(unittest.TestCase):
    def test_exception_can_include_message_string(self):
        exc = exceptions.MopidyException("foo")

        assert exc.message == "foo"
        assert str(exc) == "foo"

    def test_backend_error_is_a_mopidy_exception(self):
        assert issubclass(exceptions.BackendError, exceptions.MopidyException)

    def test_extension_error_is_a_mopidy_exception(self):
        assert issubclass(exceptions.ExtensionError, exceptions.MopidyException)

    def test_frontend_error_is_a_mopidy_exception(self):
        assert issubclass(exceptions.FrontendError, exceptions.MopidyException)

    def test_mixer_error_is_a_mopidy_exception(self):
        assert issubclass(exceptions.MixerError, exceptions.MopidyException)

    def test_scanner_error_is_a_mopidy_exception(self):
        assert issubclass(exceptions.ScannerError, exceptions.MopidyException)

    def test_audio_error_is_a_mopidy_exception(self):
        assert issubclass(exceptions.AudioException, exceptions.MopidyException)



import unittest
from unittest import mock

from mopidy import backend, core
from mopidy.models import Playlist, Ref, Track


class BasePlaylistsTest(unittest.TestCase):
    def setUp(self):  # noqa: N802
        self.plr1a = Ref.playlist(name="A", uri="dummy1:pl:a")
        self.plr1b = Ref.playlist(name="B", uri="dummy1:pl:b")
        self.plr2a = Ref.playlist(name="A", uri="dummy2:pl:a")
        self.plr2b = Ref.playlist(name="B", uri="dummy2:pl:b")

        self.pl1a = Playlist(name="A", tracks=[Track(uri="dummy1:t:a")])
        self.pl1b = Playlist(name="B", tracks=[Track(uri="dummy1:t:b")])
        self.pl2a = Playlist(name="A", tracks=[Track(uri="dummy2:t:a")])
        self.pl2b = Playlist(name="B", tracks=[Track(uri="dummy2:t:b")])

        self.sp1 = mock.Mock(spec=backend.PlaylistsProvider)
        self.sp1.as_list.return_value.get.return_value = [
            self.plr1a,
            self.plr1b,
        ]
        self.sp1.lookup.return_value.get.side_effect = [self.pl1a, self.pl1b]

        self.sp2 = mock.Mock(spec=backend.PlaylistsProvider)
        self.sp2.as_list.return_value.get.return_value = [
            self.plr2a,
            self.plr2b,
        ]
        self.sp2.lookup.return_value.get.side_effect = [self.pl2a, self.pl2b]

        self.backend1 = mock.Mock()
        self.backend1.actor_ref.actor_class.__name__ = "Backend1"
        self.backend1.uri_schemes.get.return_value = ["dummy1"]
        self.backend1.playlists = self.sp1

        self.backend2 = mock.Mock()
        self.backend2.actor_ref.actor_class.__name__ = "Backend2"
        self.backend2.uri_schemes.get.return_value = ["dummy2"]
        self.backend2.playlists = self.sp2

        # A backend without the optional playlists provider
        self.backend3 = mock.Mock()
        self.backend3.uri_schemes.get.return_value = ["dummy3"]
        self.backend3.has_playlists().get.return_value = False
        self.backend3.playlists = None

        self.core = core.Core(
            mixer=None, backends=[self.backend3, self.backend1, self.backend2]
        )


class PlaylistTest(BasePlaylistsTest):
    def test_as_list_combines_result_from_backends(self):
        result = self.core.playlists.as_list()

        assert self.plr1a in result
        assert self.plr1b in result
        assert self.plr2a in result
        assert self.plr2b in result

    def test_as_list_ignores_backends_that_dont_support_it(self):
        self.sp2.as_list.return_value.get.side_effect = NotImplementedError

        result = self.core.playlists.as_list()

        assert len(result) == 2
        assert self.plr1a in result
        assert self.plr1b in result

    def test_get_items_selects_the_matching_backend(self):
        ref = Ref.track()
        self.sp2.get_items.return_value.get.return_value = [ref]

        result = self.core.playlists.get_items("dummy2:pl:a")

        assert [ref] == result
        assert not self.sp1.get_items.called
        self.sp2.get_items.assert_called_once_with("dummy2:pl:a")

    def test_get_items_with_unknown_uri_scheme_does_nothing(self):
        result = self.core.playlists.get_items("unknown:a")

        assert result is None
        assert not self.sp1.delete.called
        assert not self.sp2.delete.called

    def test_create_without_uri_scheme_uses_first_backend(self):
        playlist = Playlist()
        self.sp1.create.return_value.get.return_value = playlist

        result = self.core.playlists.create("foo")

        assert playlist == result
        self.sp1.create.assert_called_once_with("foo")
        assert not self.sp2.create.called

    def test_create_without_uri_scheme_ignores_none_result(self):
        playlist = Playlist()
        self.sp1.create.return_value.get.return_value = None
        self.sp2.create.return_value.get.return_value = playlist

        result = self.core.playlists.create("foo")

        assert playlist == result
        self.sp1.create.assert_called_once_with("foo")
        self.sp2.create.assert_called_once_with("foo")

    def test_create_without_uri_scheme_ignores_exception(self):
        playlist = Playlist()
        self.sp1.create.return_value.get.side_effect = Exception
        self.sp2.create.return_value.get.return_value = playlist

        result = self.core.playlists.create("foo")

        assert playlist == result
        self.sp1.create.assert_called_once_with("foo")
        self.sp2.create.assert_called_once_with("foo")

    def test_create_with_uri_scheme_selects_the_matching_backend(self):
        playlist = Playlist()
        self.sp2.create.return_value.get.return_value = playlist

        result = self.core.playlists.create("foo", uri_scheme="dummy2")

        assert playlist == result
        assert not self.sp1.create.called
        self.sp2.create.assert_called_once_with("foo")

    def test_create_with_unsupported_uri_scheme_uses_first_backend(self):
        playlist = Playlist()
        self.sp1.create.return_value.get.return_value = playlist

        result = self.core.playlists.create("foo", uri_scheme="dummy3")

        assert playlist == result
        self.sp1.create.assert_called_once_with("foo")
        assert not self.sp2.create.called

    def test_delete_selects_the_dummy1_backend(self):
        success = self.core.playlists.delete("dummy1:a")

        assert success
        self.sp1.delete.assert_called_once_with("dummy1:a")
        assert not self.sp2.delete.called

    def test_delete_selects_the_dummy2_backend(self):
        success = self.core.playlists.delete("dummy2:a")

        assert success
        assert not self.sp1.delete.called
        self.sp2.delete.assert_called_once_with("dummy2:a")

    def test_delete_with_unknown_uri_scheme_does_nothing(self):
        success = self.core.playlists.delete("unknown:a")

        assert not success
        assert not self.sp1.delete.called
        assert not self.sp2.delete.called

    def test_delete_ignores_backend_without_playlist_support(self):
        success = self.core.playlists.delete("dummy3:a")

        assert not success
        assert not self.sp1.delete.called
        assert not self.sp2.delete.called

    def test_lookup_selects_the_dummy1_backend(self):
        self.core.playlists.lookup("dummy1:a")

        self.sp1.lookup.assert_called_once_with("dummy1:a")
        assert not self.sp2.lookup.called

    def test_lookup_selects_the_dummy2_backend(self):
        self.core.playlists.lookup("dummy2:a")

        assert not self.sp1.lookup.called
        self.sp2.lookup.assert_called_once_with("dummy2:a")

    def test_lookup_track_in_backend_without_playlists_fails(self):
        result = self.core.playlists.lookup("dummy3:a")

        assert result is None
        assert not self.sp1.lookup.called
        assert not self.sp2.lookup.called

    def test_refresh_without_uri_scheme_refreshes_all_backends(self):
        self.core.playlists.refresh()

        self.sp1.refresh.assert_called_once_with()
        self.sp2.refresh.assert_called_once_with()

    def test_refresh_with_uri_scheme_refreshes_matching_backend(self):
        self.core.playlists.refresh(uri_scheme="dummy2")

        assert not self.sp1.refresh.called
        self.sp2.refresh.assert_called_once_with()

    def test_refresh_with_unknown_uri_scheme_refreshes_nothing(self):
        self.core.playlists.refresh(uri_scheme="foobar")

        assert not self.sp1.refresh.called
        assert not self.sp2.refresh.called

    def test_refresh_ignores_backend_without_playlist_support(self):
        self.core.playlists.refresh(uri_scheme="dummy3")

        assert not self.sp1.refresh.called
        assert not self.sp2.refresh.called

    def test_save_selects_the_dummy1_backend(self):
        playlist = Playlist(uri="dummy1:a")
        self.sp1.save.return_value.get.return_value = playlist

        result = self.core.playlists.save(playlist)

        assert playlist == result
        self.sp1.save.assert_called_once_with(playlist)
        assert not self.sp2.save.called

    def test_save_selects_the_dummy2_backend(self):
        playlist = Playlist(uri="dummy2:a")
        self.sp2.save.return_value.get.return_value = playlist

        result = self.core.playlists.save(playlist)

        assert playlist == result
        assert not self.sp1.save.called
        self.sp2.save.assert_called_once_with(playlist)

    def test_save_does_nothing_if_playlist_uri_is_unset(self):
        result = self.core.playlists.save(Playlist())

        assert result is None
        assert not self.sp1.save.called
        assert not self.sp2.save.called

    def test_save_does_nothing_if_playlist_uri_has_unknown_scheme(self):
        result = self.core.playlists.save(Playlist(uri="foobar:a"))

        assert result is None
        assert not self.sp1.save.called
        assert not self.sp2.save.called

    def test_save_ignores_backend_without_playlist_support(self):
        result = self.core.playlists.save(Playlist(uri="dummy3:a"))

        assert result is None
        assert not self.sp1.save.called
        assert not self.sp2.save.called

    def test_get_uri_schemes(self):
        result = self.core.playlists.get_uri_schemes()
        assert result == ["dummy1", "dummy2"]


class MockBackendCorePlaylistsBase(unittest.TestCase):
    def setUp(self):  # noqa: N802
        self.playlists = mock.Mock(spec=backend.PlaylistsProvider)

        self.backend = mock.Mock()
        self.backend.actor_ref.actor_class.__name__ = "DummyBackend"
        self.backend.uri_schemes.get.return_value = ["dummy"]
        self.backend.playlists = self.playlists

        self.core = core.Core(mixer=None, backends=[self.backend])


@mock.patch("mopidy.core.playlists.logger")
class AsListBadBackendsTest(MockBackendCorePlaylistsBase):
    def test_backend_raises_exception(self, logger):
        self.playlists.as_list.return_value.get.side_effect = Exception
        assert [] == self.core.playlists.as_list()
        logger.exception.assert_called_with(mock.ANY, "DummyBackend")

    def test_backend_returns_none(self, logger):
        self.playlists.as_list.return_value.get.return_value = None
        assert [] == self.core.playlists.as_list()
        assert not logger.error.called

    def test_backend_returns_wrong_type(self, logger):
        self.playlists.as_list.return_value.get.return_value = "abc"
        assert [] == self.core.playlists.as_list()
        logger.error.assert_called_with(mock.ANY, "DummyBackend", mock.ANY)


@mock.patch("mopidy.core.playlists.logger")
class GetItemsBadBackendsTest(MockBackendCorePlaylistsBase):
    def test_backend_raises_exception(self, logger):
        self.playlists.get_items.return_value.get.side_effect = Exception
        assert self.core.playlists.get_items("dummy:/1") is None
        logger.exception.assert_called_with(mock.ANY, "DummyBackend")

    def test_backend_returns_none(self, logger):
        self.playlists.get_items.return_value.get.return_value = None
        assert self.core.playlists.get_items("dummy:/1") is None
        assert not logger.error.called

    def test_backend_returns_wrong_type(self, logger):
        self.playlists.get_items.return_value.get.return_value = "abc"
        assert self.core.playlists.get_items("dummy:/1") is None
        logger.error.assert_called_with(mock.ANY, "DummyBackend", mock.ANY)


@mock.patch("mopidy.core.playlists.logger")
class CreateBadBackendsTest(MockBackendCorePlaylistsBase):
    def test_backend_raises_exception(self, logger):
        self.playlists.create.return_value.get.side_effect = Exception
        assert self.core.playlists.create("foobar") is None
        logger.exception.assert_called_with(mock.ANY, "DummyBackend")

    def test_backend_returns_none(self, logger):
        self.playlists.create.return_value.get.return_value = None
        assert self.core.playlists.create("foobar") is None
        assert not logger.error.called

    def test_backend_returns_wrong_type(self, logger):
        self.playlists.create.return_value.get.return_value = "abc"
        assert self.core.playlists.create("foobar") is None
        logger.error.assert_called_with(mock.ANY, "DummyBackend", mock.ANY)


@mock.patch("mopidy.core.playlists.logger")
class DeleteBadBackendsTest(MockBackendCorePlaylistsBase):
    def test_backend_raises_exception(self, logger):
        self.playlists.delete.return_value.get.side_effect = Exception
        assert not self.core.playlists.delete("dummy:/1")
        logger.exception.assert_called_with(mock.ANY, "DummyBackend")


@mock.patch("mopidy.core.playlists.logger")
class LookupBadBackendsTest(MockBackendCorePlaylistsBase):
    def test_backend_raises_exception(self, logger):
        self.playlists.lookup.return_value.get.side_effect = Exception
        assert self.core.playlists.lookup("dummy:/1") is None
        logger.exception.assert_called_with(mock.ANY, "DummyBackend")

    def test_backend_returns_none(self, logger):
        self.playlists.lookup.return_value.get.return_value = None
        assert self.core.playlists.lookup("dummy:/1") is None
        assert not logger.error.called

    def test_backend_returns_wrong_type(self, logger):
        self.playlists.lookup.return_value.get.return_value = "abc"
        assert self.core.playlists.lookup("dummy:/1") is None
        logger.error.assert_called_with(mock.ANY, "DummyBackend", mock.ANY)


@mock.patch("mopidy.core.playlists.logger")
class RefreshBadBackendsTest(MockBackendCorePlaylistsBase):
    @mock.patch("mopidy.core.listener.CoreListener.send")
    def test_backend_raises_exception(self, send, logger):
        self.playlists.refresh.return_value.get.side_effect = Exception
        self.core.playlists.refresh()
        assert not send.called
        logger.exception.assert_called_with(mock.ANY, "DummyBackend")

    @mock.patch("mopidy.core.listener.CoreListener.send")
    def test_backend_raises_exception_called_with_uri(self, send, logger):
        self.playlists.refresh.return_value.get.side_effect = Exception
        self.core.playlists.refresh("dummy")
        assert not send.called
        logger.exception.assert_called_with(mock.ANY, "DummyBackend")


@mock.patch("mopidy.core.playlists.logger")
class SaveBadBackendsTest(MockBackendCorePlaylistsBase):
    def test_backend_raises_exception(self, logger):
        playlist = Playlist(uri="dummy:/1")
        self.playlists.save.return_value.get.side_effect = Exception
        assert self.core.playlists.save(playlist) is None
        logger.exception.assert_called_with(mock.ANY, "DummyBackend")

    def test_backend_returns_none(self, logger):
        playlist = Playlist(uri="dummy:/1")
        self.playlists.save.return_value.get.return_value = None
        assert self.core.playlists.save(playlist) is None
        assert not logger.error.called

    def test_backend_returns_wrong_type(self, logger):
        playlist = Playlist(uri="dummy:/1")
        self.playlists.save.return_value.get.return_value = "abc"
        assert self.core.playlists.save(playlist) is None
        logger.error.assert_called_with(mock.ANY, "DummyBackend", mock.ANY)



import pykka

from mopidy import backend

from . import playlists


class M3UBackend(pykka.ThreadingActor, backend.Backend):
    uri_schemes = ["m3u"]

    def __init__(self, config, audio):
        super().__init__()
        self.playlists = playlists.M3UPlaylistsProvider(self, config)



import logging
import socket
from unittest import mock

import pytest

from mopidy.config import types
from mopidy.internal import log


@pytest.mark.parametrize(
    "value, expected",
    [
        # bytes are coded from UTF-8 and string-escaped:
        (b"abc", "abc"),
        ("æøå".encode(), "æøå"),
        (b"a\nb", "a\nb"),
        (b"a\\nb", "a\nb"),
        # unicode strings are string-escaped:
        ("abc", "abc"),
        ("æøå", "æøå"),
        ("a\nb", "a\nb"),
        ("a\\nb", "a\nb"),
    ],
)
def test_decode(value, expected):
    assert types.decode(value) == expected


@pytest.mark.parametrize(
    "value, expected",
    [
        # unicode strings are string-escaped and encoded as UTF-8:
        ("abc", "abc"),
        ("æøå", "æøå"),
        ("a\nb", "a\\nb"),
        # bytes are string-escaped:
        (b"abc", "abc"),
        (b"a\nb", "a\\nb"),
    ],
)
def test_encode(value, expected):
    assert types.encode(value) == expected


def test_encode_decode_invalid_utf8():
    data = b"\xc3\x00"  # invalid utf-8

    result = types.encode(types.decode(data))

    assert isinstance(result, str)
    assert result == data.decode(errors="surrogateescape")


class TestConfigValue:
    def test_deserialize_decodes_bytes(self):
        cv = types.ConfigValue()

        result = cv.deserialize(b"abc")

        assert isinstance(result, str)

    def test_serialize_conversion_to_string(self):
        cv = types.ConfigValue()

        result = cv.serialize(object())

        assert isinstance(result, str)

    def test_serialize_none(self):
        cv = types.ConfigValue()

        result = cv.serialize(None)

        assert isinstance(result, str)
        assert result == ""

    def test_serialize_supports_display(self):
        cv = types.ConfigValue()

        result = cv.serialize(object(), display=True)

        assert isinstance(result, str)


class TestDeprecated:
    def test_deserialize_returns_deprecated_value(self):
        cv = types.Deprecated()

        result = cv.deserialize(b"foobar")

        assert isinstance(result, types.DeprecatedValue)

    def test_serialize_returns_deprecated_value(self):
        cv = types.Deprecated()

        result = cv.serialize("foobar")

        assert isinstance(result, types.DeprecatedValue)


class TestString:
    def test_deserialize_conversion_success(self):
        cv = types.String()

        result = cv.deserialize(b" foo ")

        assert result == "foo"
        assert isinstance(result, str)

    def test_deserialize_decodes_utf8(self):
        cv = types.String()

        result = cv.deserialize("æøå".encode())

        assert result == "æøå"

    def test_deserialize_does_not_double_encode_unicode(self):
        cv = types.String()

        result = cv.deserialize("æøå")

        assert result == "æøå"

    def test_deserialize_handles_escapes(self):
        cv = types.String(optional=True)

        result = cv.deserialize(b"a\\t\\nb")

        assert result == "a\t\nb"

    def test_deserialize_enforces_choices(self):
        cv = types.String(choices=["foo", "bar", "baz"])

        assert cv.deserialize(b"foo") == "foo"

        with pytest.raises(ValueError):
            cv.deserialize(b"foobar")

    def test_deserialize_enforces_required(self):
        cv = types.String()

        with pytest.raises(ValueError):
            cv.deserialize(b"")

    def test_deserialize_respects_optional(self):
        cv = types.String(optional=True)

        assert cv.deserialize(b"") is None
        assert cv.deserialize(b" ") is None

    def test_deserialize_invalid_encoding(self):
        cv = types.String()
        incorrectly_encoded_bytes = "æøå".encode("iso-8859-1")

        assert cv.deserialize(incorrectly_encoded_bytes) == "\udce6\udcf8\udce5"

    def test_serialize_returns_text(self):
        cv = types.String()

        result = cv.serialize("æøå")

        assert isinstance(result, str)
        assert result == "æøå"

    def test_serialize_decodes_bytes(self):
        cv = types.String()
        bytes_string = "æøå".encode()

        result = cv.serialize(bytes_string)

        assert isinstance(result, str)
        assert result == bytes_string.decode()

    def test_serialize_handles_escapes(self):
        cv = types.String()

        result = cv.serialize("a\n\tb")

        assert isinstance(result, str)
        assert result == r"a\n\tb"

    def test_serialize_none(self):
        cv = types.String()

        result = cv.serialize(None)

        assert isinstance(result, str)
        assert result == ""

    def test_deserialize_enforces_choices_optional(self):
        cv = types.String(optional=True, choices=["foo", "bar", "baz"])

        assert cv.deserialize(b"") is None

        with pytest.raises(ValueError):
            cv.deserialize(b"foobar")


class TestSecret:
    def test_deserialize_decodes_utf8(self):
        cv = types.Secret()

        result = cv.deserialize("æøå".encode())

        assert isinstance(result, str)
        assert result == "æøå"

    def test_deserialize_enforces_required(self):
        cv = types.Secret()

        with pytest.raises(ValueError):
            cv.deserialize(b"")

    def test_deserialize_respects_optional(self):
        cv = types.Secret(optional=True)

        assert cv.deserialize(b"") is None
        assert cv.deserialize(b" ") is None

    def test_serialize_none(self):
        cv = types.Secret()

        result = cv.serialize(None)

        assert isinstance(result, str)
        assert result == ""

    def test_serialize_for_display_masks_value(self):
        cv = types.Secret()

        result = cv.serialize("s3cret", display=True)

        assert isinstance(result, str)
        assert result == "********"

    def test_serialize_none_for_display(self):
        cv = types.Secret()

        result = cv.serialize(None, display=True)

        assert isinstance(result, str)
        assert result == ""


class TestInteger:
    def test_deserialize_conversion_success(self):
        cv = types.Integer()

        assert cv.deserialize("123") == 123
        assert cv.deserialize("0") == 0
        assert cv.deserialize("-10") == -10

    def test_deserialize_conversion_failure(self):
        cv = types.Integer()

        with pytest.raises(ValueError):
            cv.deserialize("asd")

        with pytest.raises(ValueError):
            cv.deserialize("3.14")

    def test_deserialize_enforces_required(self):
        cv = types.Integer()

        with pytest.raises(ValueError):
            cv.deserialize("")

    def test_deserialize_respects_optional(self):
        cv = types.Integer(optional=True)

        assert cv.deserialize("") is None

    def test_deserialize_enforces_choices(self):
        cv = types.Integer(choices=[1, 2, 3])

        assert cv.deserialize("3") == 3

        with pytest.raises(ValueError):
            cv.deserialize("5")

    def test_deserialize_enforces_minimum(self):
        cv = types.Integer(minimum=10)

        assert cv.deserialize("15") == 15

        with pytest.raises(ValueError):
            cv.deserialize("5")

    def test_deserialize_enforces_maximum(self):
        cv = types.Integer(maximum=10)

        assert cv.deserialize("5") == 5

        with pytest.raises(ValueError):
            cv.deserialize("15")


class TestBoolean:
    def test_deserialize_conversion_success(self):
        cv = types.Boolean()

        for true in ("1", "yes", "true", "on"):
            assert cv.deserialize(true) is True
            assert cv.deserialize(true.upper()) is True
            assert cv.deserialize(true.capitalize()) is True

        for false in ("0", "no", "false", "off"):
            assert cv.deserialize(false) is False
            assert cv.deserialize(false.upper()) is False
            assert cv.deserialize(false.capitalize()) is False

    def test_deserialize_conversion_failure(self):
        cv = types.Boolean()

        with pytest.raises(ValueError):
            cv.deserialize("nope")

        with pytest.raises(ValueError):
            cv.deserialize("sure")

    def test_deserialize_enforces_required(self):
        cv = types.Boolean()

        with pytest.raises(ValueError):
            cv.deserialize("")

    def test_deserialize_respects_optional(self):
        cv = types.Boolean(optional=True)

        assert cv.deserialize("") is None

    def test_serialize_true(self):
        cv = types.Boolean()

        result = cv.serialize(True)

        assert isinstance(result, str)
        assert result == "true"

    def test_serialize_false(self):
        cv = types.Boolean()

        result = cv.serialize(False)

        assert isinstance(result, str)
        assert result == "false"

    def test_serialize_none_as_false(self):
        # TODO We should consider making `None` an invalid value, but we have
        # existing code that assumes it to work like False.

        cv = types.Boolean()

        result = cv.serialize(None)

        assert isinstance(result, str)
        assert result == "false"

    def test_serialize_invalid_values(self):
        cv = types.Boolean()

        with pytest.raises(ValueError):
            cv.serialize([])

        with pytest.raises(ValueError):
            cv.serialize("1")


class TestList:
    # TODO: add test_deserialize_ignores_blank
    # TODO: add test_serialize_ignores_blank
    # TODO: add test_deserialize_handles_escapes

    def test_deserialize_conversion_success(self):
        cv = types.List()

        result = cv.deserialize(b"foo, bar ,baz ")
        assert result == ("foo", "bar", "baz")

        result = cv.deserialize(b" foo,bar\nbar\nbaz")
        assert result == ("foo,bar", "bar", "baz")

    def test_deserialize_creates_tuples(self):
        cv = types.List(optional=True)

        assert isinstance(cv.deserialize(b"foo,bar,baz"), tuple)
        assert isinstance(cv.deserialize(b""), tuple)

    def test_deserialize_decodes_utf8(self):
        cv = types.List()

        result = cv.deserialize("æ, ø, å".encode())
        assert result == ("æ", "ø", "å")

        result = cv.deserialize("æ\nø\nå".encode())
        assert result == ("æ", "ø", "å")

    def test_deserialize_does_not_double_encode_unicode(self):
        cv = types.List()

        result = cv.deserialize("æ, ø, å")
        assert result == ("æ", "ø", "å")

        result = cv.deserialize("æ\nø\nå")
        assert result == ("æ", "ø", "å")

    def test_deserialize_enforces_required(self):
        cv = types.List()

        with pytest.raises(ValueError):
            cv.deserialize(b"")

    def test_deserialize_respects_optional(self):
        cv = types.List(optional=True)

        assert cv.deserialize(b"") == ()

    def test_serialize(self):
        cv = types.List()

        result = cv.serialize(("foo", "bar", "baz"))

        assert isinstance(result, str)
        assert result == "\n  foo\n  bar\n  baz"

    def test_serialize_none(self):
        cv = types.List()

        result = cv.serialize(None)

        assert isinstance(result, str)
        assert result == ""


class TestLogColor:
    def test_deserialize(self):
        cv = types.LogColor()

        assert cv.deserialize(b"RED") == "red"
        assert cv.deserialize("RED") == "red"
        assert cv.deserialize(b"red") == "red"
        assert cv.deserialize("red") == "red"

    def test_deserialize_enforces_choices(self):
        cv = types.LogColor()

        with pytest.raises(ValueError):
            cv.deserialize("golden")

    def test_serialize(self):
        cv = types.LogColor()

        assert cv.serialize("red") == "red"
        assert cv.serialize("blue") == "blue"

    def test_serialize_ignores_unknown_color(self):
        cv = types.LogColor()

        assert cv.serialize("golden") == ""


class TestLogLevel:
    levels = {
        "critical": logging.CRITICAL,
        "error": logging.ERROR,
        "warning": logging.WARNING,
        "info": logging.INFO,
        "debug": logging.DEBUG,
        "trace": log.TRACE_LOG_LEVEL,
        "all": logging.NOTSET,
    }

    def test_deserialize_conversion_success(self):
        cv = types.LogLevel()

        for name, level in self.levels.items():
            assert cv.deserialize(name) == level
            assert cv.deserialize(name.upper()) == level
            assert cv.deserialize(name.capitalize()) == level

    def test_deserialize_conversion_failure(self):
        cv = types.LogLevel()

        with pytest.raises(ValueError):
            cv.deserialize("red alert")

        with pytest.raises(ValueError):
            cv.deserialize(" ")

    def test_serialize(self):
        cv = types.LogLevel()

        for name, level in self.levels.items():
            assert cv.serialize(level) == name

    def test_serialize_ignores_unknown_level(self):
        cv = types.LogLevel()

        assert cv.serialize(1337) == ""


class TestHostname:
    @mock.patch("socket.getaddrinfo")
    def test_deserialize_conversion_success(self, getaddrinfo_mock):
        cv = types.Hostname()

        cv.deserialize("example.com")

        getaddrinfo_mock.assert_called_once_with("example.com", None)

    @mock.patch("socket.getaddrinfo")
    def test_deserialize_conversion_failure(self, getaddrinfo_mock):
        cv = types.Hostname()
        getaddrinfo_mock.side_effect = socket.error

        with pytest.raises(ValueError):
            cv.deserialize("example.com")

    @mock.patch("socket.getaddrinfo")
    def test_deserialize_enforces_required(self, getaddrinfo_mock):
        cv = types.Hostname()

        with pytest.raises(ValueError):
            cv.deserialize("")

        with pytest.raises(ValueError):
            cv.deserialize(" ")

        assert getaddrinfo_mock.call_count == 0

    @mock.patch("socket.getaddrinfo")
    def test_deserialize_respects_optional(self, getaddrinfo_mock):
        cv = types.Hostname(optional=True)

        assert cv.deserialize("") is None
        assert cv.deserialize(" ") is None
        assert getaddrinfo_mock.call_count == 0

    @mock.patch("mopidy.internal.path.expand_path")
    def test_deserialize_with_unix_socket(self, expand_path_mock):
        cv = types.Hostname()

        assert cv.deserialize("unix:/tmp/mopidy.socket") is not None
        expand_path_mock.assert_called_once_with("/tmp/mopidy.socket")


class TestPort:
    def test_valid_ports(self):
        cv = types.Port()

        assert cv.deserialize("0") == 0
        assert cv.deserialize("1") == 1
        assert cv.deserialize("80") == 80
        assert cv.deserialize("6600") == 6600
        assert cv.deserialize("65535") == 65535

    def test_invalid_ports(self):
        cv = types.Port()

        with pytest.raises(ValueError):
            cv.deserialize("65536")

        with pytest.raises(ValueError):
            cv.deserialize("100000")

        with pytest.raises(ValueError):
            cv.deserialize("-1")

        with pytest.raises(ValueError):
            cv.deserialize("")


class TestExpandedPath:
    def test_is_str(self):
        assert isinstance(types._ExpandedPath(b"/tmp", b"foo"), str)

    def test_stores_both_expanded_and_original_path(self):
        original = "~"
        expanded = "expanded_path"

        result = types._ExpandedPath(original, expanded)

        assert result == expanded
        assert result.original == original


class TestPath:
    def test_deserialize_conversion_success(self):
        cv = types.Path()

        result = cv.deserialize(b"/foo")

        assert result == "/foo"
        assert isinstance(result, types._ExpandedPath)
        assert isinstance(result, str)

    def test_deserialize_enforces_required(self):
        cv = types.Path()

        with pytest.raises(ValueError):
            cv.deserialize(b"")

    def test_deserialize_respects_optional(self):
        cv = types.Path(optional=True)

        assert cv.deserialize(b"") is None
        assert cv.deserialize(b" ") is None

    def test_serialize_uses_original(self):
        cv = types.Path()
        path = types._ExpandedPath("original_path", "expanded_path")

        assert cv.serialize(path) == "original_path"

    def test_serialize_plain_string(self):
        cv = types.Path()

        assert cv.serialize(b"path") == "path"

    def test_serialize_supports_unicode_string(self):
        cv = types.Path()

        assert cv.serialize("æøå") == "æøå"



import unittest

from mopidy import exceptions
from mopidy.audio import scan
from mopidy.internal.path import path_to_uri

from tests import path_to_data_dir


class ScannerTest(unittest.TestCase):
    def setUp(self):  # noqa: N802
        self.errors = {}
        self.result = {}

    def find(self, path):
        dir_path = path_to_data_dir(path)
        if not dir_path.is_dir():
            return
        for file_path in dir_path.iterdir():
            yield dir_path / file_path

    def scan(self, paths):
        scanner = scan.Scanner()
        for path in paths:
            uri = path_to_uri(path)
            try:
                self.result[path] = scanner.scan(uri)
            except exceptions.ScannerError as error:
                self.errors[path] = error

    def check(self, name, key, value):
        name = path_to_data_dir(name)
        assert self.result[name].tags[key] == value

    def check_if_missing_plugin(self):
        for path, result in self.result.items():
            if path.suffix != ".mp3":
                continue
            if not result.playable and result.mime == "audio/mpeg":
                raise unittest.SkipTest("Missing MP3 support?")

    def test_tags_is_set(self):
        self.scan(self.find("scanner/simple"))

        assert list(self.result.values())[0].tags

    def test_errors_is_not_set(self):
        self.scan(self.find("scanner/simple"))

        self.check_if_missing_plugin()

        assert not self.errors

    def test_duration_is_set(self):
        self.scan(self.find("scanner/simple"))

        self.check_if_missing_plugin()

        ogg = path_to_data_dir("scanner/simple/song1.ogg")
        mp3 = path_to_data_dir("scanner/simple/song1.mp3")
        assert self.result[mp3].duration == 4680
        assert self.result[ogg].duration == 4680

    def test_artist_is_set(self):
        self.scan(self.find("scanner/simple"))

        self.check_if_missing_plugin()

        self.check("scanner/simple/song1.mp3", "artist", ["name"])
        self.check("scanner/simple/song1.ogg", "artist", ["name"])

    def test_album_is_set(self):
        self.scan(self.find("scanner/simple"))

        self.check_if_missing_plugin()

        self.check("scanner/simple/song1.mp3", "album", ["albumname"])
        self.check("scanner/simple/song1.ogg", "album", ["albumname"])

    def test_track_is_set(self):
        self.scan(self.find("scanner/simple"))

        self.check_if_missing_plugin()

        self.check("scanner/simple/song1.mp3", "title", ["trackname"])
        self.check("scanner/simple/song1.ogg", "title", ["trackname"])

    def test_nonexistant_dir_does_not_fail(self):
        self.scan(self.find("scanner/does-not-exist"))
        assert not self.errors

    def test_other_media_is_ignored(self):
        self.scan(self.find("scanner/image"))
        assert not list(self.result.values())[0].playable

    def test_log_file_that_gst_thinks_is_mpeg_1_is_ignored(self):
        self.scan([path_to_data_dir("scanner/example.log")])

        self.check_if_missing_plugin()

        log = path_to_data_dir("scanner/example.log")
        assert self.result[log].duration is None

    def test_empty_wav_file(self):
        self.scan([path_to_data_dir("scanner/empty.wav")])
        wav = path_to_data_dir("scanner/empty.wav")
        assert self.result[wav].duration == 0

    def test_uri_list(self):
        path = path_to_data_dir("scanner/playlist.m3u")
        self.scan([path])
        assert self.result[path].mime == "text/uri-list"

    def test_text_plain(self):
        # GStreamer decode bin hardcodes bad handling of text plain :/
        path = path_to_data_dir("scanner/plain.txt")
        self.scan([path])
        assert path in self.errors

    @unittest.SkipTest
    def test_song_without_time_is_handeled(self):
        pass



import argparse
import unittest
from unittest import mock

from mopidy import commands


class ConfigOverrideTypeTest(unittest.TestCase):
    def test_valid_override(self):
        expected = ("section", "key", "value")
        assert expected == commands.config_override_type("section/key=value")
        assert expected == commands.config_override_type("section/key=value ")
        assert expected == commands.config_override_type("section/key =value")
        assert expected == commands.config_override_type("section /key=value")

    def test_empty_override(self):
        expected = ("section", "key", "")
        assert expected == commands.config_override_type("section/key=")
        assert expected == commands.config_override_type("section/key=  ")

    def test_invalid_override(self):
        with self.assertRaises(argparse.ArgumentTypeError):
            commands.config_override_type("section/key")
        with self.assertRaises(argparse.ArgumentTypeError):
            commands.config_override_type("section=")
        with self.assertRaises(argparse.ArgumentTypeError):
            commands.config_override_type("section")


class CommandParsingTest(unittest.TestCase):
    def setUp(self):  # noqa: N802
        self.exit_patcher = mock.patch.object(commands.Command, "exit")
        self.exit_mock = self.exit_patcher.start()
        self.exit_mock.side_effect = SystemExit

    def tearDown(self):  # noqa: N802
        self.exit_patcher.stop()

    def test_command_parsing_returns_namespace(self):
        cmd = commands.Command()
        assert isinstance(cmd.parse([]), argparse.Namespace)

    def test_command_parsing_does_not_contain_args(self):
        cmd = commands.Command()
        result = cmd.parse([])
        assert not hasattr(result, "_args")

    def test_unknown_options_bails(self):
        cmd = commands.Command()
        with self.assertRaises(SystemExit):
            cmd.parse(["--foobar"])

    def test_invalid_sub_command_bails(self):
        cmd = commands.Command()
        with self.assertRaises(SystemExit):
            cmd.parse(["foo"])

    def test_command_arguments(self):
        cmd = commands.Command()
        cmd.add_argument("--bar")

        result = cmd.parse(["--bar", "baz"])
        assert result.bar == "baz"

    def test_command_arguments_and_sub_command(self):
        child = commands.Command()
        child.add_argument("--baz")

        cmd = commands.Command()
        cmd.add_argument("--bar")
        cmd.add_child("foo", child)

        result = cmd.parse(["--bar", "baz", "foo"])
        assert result.bar == "baz"
        assert result.baz is None

    def test_subcommand_may_have_positional(self):
        child = commands.Command()
        child.add_argument("bar")

        cmd = commands.Command()
        cmd.add_child("foo", child)

        result = cmd.parse(["foo", "baz"])
        assert result.bar == "baz"

    def test_subcommand_may_have_remainder(self):
        child = commands.Command()
        child.add_argument("bar", nargs=argparse.REMAINDER)

        cmd = commands.Command()
        cmd.add_child("foo", child)

        result = cmd.parse(["foo", "baz", "bep", "bop"])
        assert result.bar == ["baz", "bep", "bop"]

    def test_result_stores_choosen_command(self):
        child = commands.Command()

        cmd = commands.Command()
        cmd.add_child("foo", child)

        result = cmd.parse(["foo"])
        assert result.command == child

        result = cmd.parse([])
        assert result.command == cmd

        child2 = commands.Command()
        cmd.add_child("bar", child2)

        subchild = commands.Command()
        child.add_child("baz", subchild)

        result = cmd.parse(["bar"])
        assert result.command == child2

        result = cmd.parse(["foo", "baz"])
        assert result.command == subchild

    def test_invalid_type(self):
        cmd = commands.Command()
        cmd.add_argument("--bar", type=int)

        with self.assertRaises(SystemExit):
            cmd.parse(["--bar", "zero"], prog="foo")

        self.exit_mock.assert_called_once_with(
            1,
            "argument --bar: invalid int value: 'zero'",
            "usage: foo [--bar BAR]",
        )

    @mock.patch("sys.argv")
    def test_command_error_usage_prog(self, argv_mock):
        argv_mock.__getitem__.return_value = "/usr/bin/foo"

        cmd = commands.Command()
        cmd.add_argument("--bar", required=True)

        with self.assertRaises(SystemExit):
            cmd.parse([])
        self.exit_mock.assert_called_once_with(
            mock.ANY, mock.ANY, "usage: foo --bar BAR"
        )

        self.exit_mock.reset_mock()
        with self.assertRaises(SystemExit):
            cmd.parse([], prog="baz")

        self.exit_mock.assert_called_once_with(
            mock.ANY, mock.ANY, "usage: baz --bar BAR"
        )

    def test_missing_required(self):
        cmd = commands.Command()
        cmd.add_argument("--bar", required=True)

        with self.assertRaises(SystemExit):
            cmd.parse([], prog="foo")

        self.exit_mock.assert_called_once_with(
            1,
            "the following arguments are required: --bar",
            "usage: foo --bar BAR",
        )

    def test_missing_positionals(self):
        cmd = commands.Command()
        cmd.add_argument("bar")

        with self.assertRaises(SystemExit):
            cmd.parse([], prog="foo")

        self.exit_mock.assert_called_once_with(
            1,
            "the following arguments are required: bar, _args",
            "usage: foo bar",
        )

    def test_missing_positionals_subcommand(self):
        child = commands.Command()
        child.add_argument("baz")

        cmd = commands.Command()
        cmd.add_child("bar", child)

        with self.assertRaises(SystemExit):
            cmd.parse(["bar"], prog="foo")

        self.exit_mock.assert_called_once_with(
            1,
            "the following arguments are required: baz, _args",
            "usage: foo bar baz",
        )

    def test_unknown_command(self):
        cmd = commands.Command()

        with self.assertRaises(SystemExit):
            cmd.parse(["--help"], prog="foo")

        self.exit_mock.assert_called_once_with(
            1, "unrecognized arguments: --help", "usage: foo"
        )

    def test_invalid_subcommand(self):
        cmd = commands.Command()
        cmd.add_child("baz", commands.Command())

        with self.assertRaises(SystemExit):
            cmd.parse(["bar"], prog="foo")

        self.exit_mock.assert_called_once_with(
            1, "unrecognized command: bar", "usage: foo"
        )

    def test_set(self):
        cmd = commands.Command()
        cmd.set(foo="bar")

        result = cmd.parse([])
        assert result.foo == "bar"

    def test_set_propegate(self):
        child = commands.Command()

        cmd = commands.Command()
        cmd.set(foo="bar")
        cmd.add_child("command", child)

        result = cmd.parse(["command"])
        assert result.foo == "bar"

    def test_innermost_set_wins(self):
        child = commands.Command()
        child.set(foo="bar", baz=1)

        cmd = commands.Command()
        cmd.set(foo="baz", baz=None)
        cmd.add_child("command", child)

        result = cmd.parse(["command"])
        assert result.foo == "bar"
        assert result.baz == 1

    def test_help_action_works(self):
        cmd = commands.Command()
        cmd.add_argument("-h", action="help")
        cmd.format_help = mock.Mock()

        with self.assertRaises(SystemExit):
            cmd.parse(["-h"])

        cmd.format_help.assert_called_once_with(mock.ANY)
        self.exit_mock.assert_called_once_with(0, cmd.format_help.return_value)


class UsageTest(unittest.TestCase):
    @mock.patch("sys.argv")
    def test_prog_name_default_and_override(self, argv_mock):
        argv_mock.__getitem__.return_value = "/usr/bin/foo"
        cmd = commands.Command()
        assert "usage: foo" == cmd.format_usage().strip()
        assert "usage: baz" == cmd.format_usage("baz").strip()

    def test_basic_usage(self):
        cmd = commands.Command()
        assert "usage: foo" == cmd.format_usage("foo").strip()

        cmd.add_argument("-h", "--help", action="store_true")
        assert "usage: foo [-h]" == cmd.format_usage("foo").strip()

        cmd.add_argument("bar")
        assert "usage: foo [-h] bar" == cmd.format_usage("foo").strip()

    def test_nested_usage(self):
        child = commands.Command()
        cmd = commands.Command()
        cmd.add_child("bar", child)

        assert "usage: foo" == cmd.format_usage("foo").strip()
        assert "usage: foo bar" == cmd.format_usage("foo bar").strip()

        cmd.add_argument("-h", "--help", action="store_true")
        assert "usage: foo bar" == child.format_usage("foo bar").strip()

        child.add_argument("-h", "--help", action="store_true")
        assert "usage: foo bar [-h]" == child.format_usage("foo bar").strip()


class HelpTest(unittest.TestCase):
    @mock.patch("sys.argv")
    def test_prog_name_default_and_override(self, argv_mock):
        argv_mock.__getitem__.return_value = "/usr/bin/foo"
        cmd = commands.Command()
        assert "usage: foo" == cmd.format_help().strip()
        assert "usage: bar" == cmd.format_help("bar").strip()

    def test_command_without_documenation_or_options(self):
        cmd = commands.Command()
        assert "usage: bar" == cmd.format_help("bar").strip()

    def test_command_with_option(self):
        cmd = commands.Command()
        cmd.add_argument(
            "-h", "--help", action="store_true", help="show this message"
        )

        expected = (
            "usage: foo [-h]\n\n"
            "OPTIONS:\n\n"
            "  -h, --help  show this message"
        )
        assert expected == cmd.format_help("foo").strip()

    def test_command_with_option_and_positional(self):
        cmd = commands.Command()
        cmd.add_argument(
            "-h", "--help", action="store_true", help="show this message"
        )
        cmd.add_argument("bar", help="some help text")

        expected = (
            "usage: foo [-h] bar\n\n"
            "OPTIONS:\n\n"
            "  -h, --help  show this message\n"
            "  bar         some help text"
        )
        assert expected == cmd.format_help("foo").strip()

    def test_command_with_documentation(self):
        cmd = commands.Command()
        cmd.help = "some text about everything this command does."

        expected = (
            "usage: foo\n\n" "some text about everything this command does."
        )
        assert expected == cmd.format_help("foo").strip()

    def test_command_with_documentation_and_option(self):
        cmd = commands.Command()
        cmd.help = "some text about everything this command does."
        cmd.add_argument(
            "-h", "--help", action="store_true", help="show this message"
        )

        expected = (
            "usage: foo [-h]\n\n"
            "some text about everything this command does.\n\n"
            "OPTIONS:\n\n"
            "  -h, --help  show this message"
        )
        assert expected == cmd.format_help("foo").strip()

    def test_subcommand_without_documentation_or_options(self):
        child = commands.Command()
        cmd = commands.Command()
        cmd.add_child("bar", child)

        assert "usage: foo" == cmd.format_help("foo").strip()

    def test_subcommand_with_documentation_shown(self):
        child = commands.Command()
        child.help = "some text about everything this command does."

        cmd = commands.Command()
        cmd.add_child("bar", child)
        expected = (
            "usage: foo\n\n"
            "COMMANDS:\n\n"
            "bar\n\n"
            "  some text about everything this command does."
        )
        assert expected == cmd.format_help("foo").strip()

    def test_subcommand_with_options_shown(self):
        child = commands.Command()
        child.add_argument(
            "-h", "--help", action="store_true", help="show this message"
        )

        cmd = commands.Command()
        cmd.add_child("bar", child)

        expected = (
            "usage: foo\n\n"
            "COMMANDS:\n\n"
            "bar [-h]\n\n"
            "    -h, --help  show this message"
        )
        assert expected == cmd.format_help("foo").strip()

    def test_subcommand_with_positional_shown(self):
        child = commands.Command()
        child.add_argument("baz", help="the great and wonderful")

        cmd = commands.Command()
        cmd.add_child("bar", child)

        expected = (
            "usage: foo\n\n"
            "COMMANDS:\n\n"
            "bar baz\n\n"
            "    baz  the great and wonderful"
        )
        assert expected == cmd.format_help("foo").strip()

    def test_subcommand_with_options_and_documentation(self):
        child = commands.Command()
        child.help = "  some text about everything this command does."
        child.add_argument(
            "-h", "--help", action="store_true", help="show this message"
        )

        cmd = commands.Command()
        cmd.add_child("bar", child)

        expected = (
            "usage: foo\n\n"
            "COMMANDS:\n\n"
            "bar [-h]\n\n"
            "  some text about everything this command does.\n\n"
            "    -h, --help  show this message"
        )
        assert expected == cmd.format_help("foo").strip()

    def test_nested_subcommands_with_options(self):
        subchild = commands.Command()
        subchild.add_argument("--test", help="the great and wonderful")

        child = commands.Command()
        child.add_child("baz", subchild)
        child.add_argument(
            "-h", "--help", action="store_true", help="show this message"
        )

        cmd = commands.Command()
        cmd.add_child("bar", child)

        expected = (
            "usage: foo\n\n"
            "COMMANDS:\n\n"
            "bar [-h]\n\n"
            "    -h, --help  show this message\n\n"
            "bar baz [--test TEST]\n\n"
            "    --test TEST  the great and wonderful"
        )
        assert expected == cmd.format_help("foo").strip()

    def test_nested_subcommands_skipped_intermediate(self):
        subchild = commands.Command()
        subchild.add_argument("--test", help="the great and wonderful")

        child = commands.Command()
        child.add_child("baz", subchild)

        cmd = commands.Command()
        cmd.add_child("bar", child)

        expected = (
            "usage: foo\n\n"
            "COMMANDS:\n\n"
            "bar baz [--test TEST]\n\n"
            "    --test TEST  the great and wonderful"
        )
        assert expected == cmd.format_help("foo").strip()

    def test_command_with_option_and_subcommand_with_option(self):
        child = commands.Command()
        child.add_argument("--test", help="the great and wonderful")

        cmd = commands.Command()
        cmd.add_argument(
            "-h", "--help", action="store_true", help="show this message"
        )
        cmd.add_child("bar", child)

        expected = (
            "usage: foo [-h]\n\n"
            "OPTIONS:\n\n"
            "  -h, --help  show this message\n\n"
            "COMMANDS:\n\n"
            "bar [--test TEST]\n\n"
            "    --test TEST  the great and wonderful"
        )
        assert expected == cmd.format_help("foo").strip()

    def test_command_with_options_doc_and_subcommand_with_option_and_doc(self):
        child = commands.Command()
        child.help = "some text about this sub-command."
        child.add_argument("--test", help="the great and wonderful")

        cmd = commands.Command()
        cmd.help = "some text about everything this command does."
        cmd.add_argument(
            "-h", "--help", action="store_true", help="show this message"
        )
        cmd.add_child("bar", child)

        expected = (
            "usage: foo [-h]\n\n"
            "some text about everything this command does.\n\n"
            "OPTIONS:\n\n"
            "  -h, --help  show this message\n\n"
            "COMMANDS:\n\n"
            "bar [--test TEST]\n\n"
            "  some text about this sub-command.\n\n"
            "    --test TEST  the great and wonderful"
        )
        assert expected == cmd.format_help("foo").strip()


class RunTest(unittest.TestCase):
    def test_default_implmentation_raises_error(self):
        with self.assertRaises(NotImplementedError):
            commands.Command().run()


class RootCommandTest(unittest.TestCase):
    def test_config_overrides(self):
        cmd = commands.RootCommand()
        result = cmd.parse(["--option", "foo/bar=baz"])

        assert result.config_overrides[0] == ("foo", "bar", "baz")



import platform
import sys
import unittest
from unittest import mock

import pkg_resources

from mopidy.internal import deps
from mopidy.internal.gi import Gst, gi


class DepsTest(unittest.TestCase):
    def test_format_dependency_list(self):
        adapters = [
            lambda: dict(name="Python", version="FooPython 2.7.3"),
            lambda: dict(name="Platform", version="Loonix 4.0.1"),
            lambda: dict(
                name="Pykka", version="1.1", path="/foo/bar", other="Quux"
            ),
            lambda: dict(name="Foo"),
            lambda: dict(
                name="Mopidy",
                version="0.13",
                dependencies=[
                    dict(
                        name="pylast",
                        version="0.5",
                        dependencies=[dict(name="setuptools", version="0.6")],
                    )
                ],
            ),
        ]

        result = deps.format_dependency_list(adapters)

        assert "Python: FooPython 2.7.3" in result

        assert "Platform: Loonix 4.0.1" in result

        assert "Pykka: 1.1 from /foo/bar" in result
        assert "/baz.py" not in result
        assert "Detailed information: Quux" in result

        assert "Foo: not found" in result

        assert "Mopidy: 0.13" in result
        assert "  pylast: 0.5" in result
        assert "    setuptools: 0.6" in result

    def test_executable_info(self):
        result = deps.executable_info()

        assert "Executable" == result["name"]
        assert sys.argv[0] in result["version"]

    def test_platform_info(self):
        result = deps.platform_info()

        assert "Platform" == result["name"]
        assert platform.platform() in result["version"]

    def test_python_info(self):
        result = deps.python_info()

        assert "Python" == result["name"]
        assert platform.python_implementation() in result["version"]
        assert platform.python_version() in result["version"]
        assert "python" in result["path"]
        assert "platform.py" not in result["path"]

    def test_gstreamer_info(self):
        result = deps.gstreamer_info()

        assert "GStreamer" == result["name"]
        assert ".".join(map(str, Gst.version())) == result["version"]
        assert "gi" in result["path"]
        assert "__init__.py" not in result["path"]
        assert "Python wrapper: python-gi" in result["other"]
        assert gi.__version__ in result["other"]
        assert "Relevant elements:" in result["other"]

    @mock.patch("pkg_resources.get_distribution")
    def test_pkg_info(self, get_distribution_mock):
        dist_setuptools = mock.Mock()
        dist_setuptools.project_name = "setuptools"
        dist_setuptools.version = "0.6"
        dist_setuptools.location = "/tmp/example/setuptools"
        dist_setuptools.requires.return_value = []

        dist_pykka = mock.Mock()
        dist_pykka.project_name = "Pykka"
        dist_pykka.version = "1.1"
        dist_pykka.location = "/tmp/example/pykka"
        dist_pykka.requires.return_value = [dist_setuptools]

        dist_mopidy = mock.Mock()
        dist_mopidy.project_name = "Mopidy"
        dist_mopidy.version = "0.13"
        dist_mopidy.location = "/tmp/example/mopidy"
        dist_mopidy.requires.return_value = [dist_pykka]

        get_distribution_mock.side_effect = [
            dist_mopidy,
            dist_pykka,
            dist_setuptools,
        ]

        result = deps.pkg_info()

        assert "Mopidy" == result["name"]
        assert "0.13" == result["version"]
        assert "mopidy" in result["path"]

        dep_info_pykka = result["dependencies"][0]
        assert "Pykka" == dep_info_pykka["name"]
        assert "1.1" == dep_info_pykka["version"]

        dep_info_setuptools = dep_info_pykka["dependencies"][0]
        assert "setuptools" == dep_info_setuptools["name"]
        assert "0.6" == dep_info_setuptools["version"]

    @mock.patch("pkg_resources.get_distribution")
    def test_pkg_info_for_missing_dist(self, get_distribution_mock):
        get_distribution_mock.side_effect = pkg_resources.DistributionNotFound

        result = deps.pkg_info()

        assert "Mopidy" == result["name"]
        assert "version" not in result
        assert "path" not in result

    @mock.patch("pkg_resources.get_distribution")
    def test_pkg_info_for_wrong_dist_version(self, get_distribution_mock):
        get_distribution_mock.side_effect = pkg_resources.VersionConflict

        result = deps.pkg_info()

        assert "Mopidy" == result["name"]
        assert "version" not in result
        assert "path" not in result



import os
import subprocess
import sys
import unittest

import mopidy


class HelpTest(unittest.TestCase):
    def test_help_has_mopidy_options(self):
        mopidy_dir = os.path.dirname(mopidy.__file__)
        args = [sys.executable, mopidy_dir, "--help"]
        process = subprocess.Popen(
            args,
            env={
                "PYTHONPATH": ":".join(
                    [
                        os.path.join(mopidy_dir, ".."),
                        os.environ.get("PYTHONPATH", ""),
                    ]
                )
            },
            stdout=subprocess.PIPE,
        )
        output = process.communicate()[0]
        assert b"--version" in output
        assert b"--help" in output
        assert b"--quiet" in output
        assert b"--verbose" in output
        assert b"--config" in output
        assert b"--option" in output



import logging
import os

import mopidy
from mopidy import config, ext

logger = logging.getLogger(__name__)


class Extension(ext.Extension):

    dist_name = "Mopidy-File"
    ext_name = "file"
    version = mopidy.__version__

    def get_default_config(self):
        conf_file = os.path.join(os.path.dirname(__file__), "ext.conf")
        return config.read(conf_file)

    def get_config_schema(self):
        schema = super().get_config_schema()
        schema["media_dirs"] = config.List(optional=True)
        schema["excluded_file_extensions"] = config.List(optional=True)
        schema["show_dotfiles"] = config.Boolean(optional=True)
        schema["follow_symlinks"] = config.Boolean(optional=True)
        schema["metadata_timeout"] = config.Integer(optional=True)
        return schema

    def setup(self, registry):
        from .backend import FileBackend

        registry.add("backend", FileBackend)



import logging
import time

import requests

from mopidy import httpclient

logger = logging.getLogger(__name__)


def get_requests_session(proxy_config, user_agent):
    proxy = httpclient.format_proxy(proxy_config)
    full_user_agent = httpclient.format_user_agent(user_agent)

    session = requests.Session()
    session.proxies.update({"http": proxy, "https": proxy})
    session.headers.update({"user-agent": full_user_agent})

    return session


def download(session, uri, timeout=1.0, chunk_size=4096):
    try:
        response = session.get(uri, stream=True, timeout=timeout)
    except requests.exceptions.Timeout:
        logger.warning(
            "Download of %r failed due to connection timeout after " "%.3fs",
            uri,
            timeout,
        )
        return None
    except requests.exceptions.InvalidSchema:
        logger.warning("Download of %r failed due to unsupported schema", uri)
        return None
    except requests.exceptions.RequestException as exc:
        logger.warning("Download of %r failed: %s", uri, exc)
        logger.debug("Download exception details", exc_info=True)
        return None

    content = []
    deadline = time.time() + timeout
    for chunk in response.iter_content(chunk_size):
        content.append(chunk)
        if time.time() > deadline:
            logger.warning(
                "Download of %r failed due to download taking more "
                "than %.3fs",
                uri,
                timeout,
            )
            return None

    if not response.ok:
        logger.warning("Problem downloading %r: %s", uri, response.reason)
        return None

    return b"".join(content)



import contextlib
import locale
import logging
import operator
import os
import pathlib
import tempfile

from mopidy import backend
from mopidy.internal import path

from . import Extension, translator

logger = logging.getLogger(__name__)


def log_environment_error(message, error):
    if isinstance(error.strerror, bytes):
        strerror = error.strerror.decode(locale.getpreferredencoding())
    else:
        strerror = error.strerror
    logger.error("%s: %s", message, strerror)


@contextlib.contextmanager
def replace(path, mode="w+b", encoding=None, errors=None):
    (fd, tempname) = tempfile.mkstemp(dir=str(path.parent))
    tempname = pathlib.Path(tempname)
    try:
        fp = open(fd, mode, encoding=encoding, errors=errors)
    except Exception:
        tempname.unlink()
        os.close(fd)
        raise
    try:
        yield fp
        fp.flush()
        os.fsync(fd)
        tempname.rename(path)
    except Exception:
        tempname.unlink()
        raise
    finally:
        fp.close()


class M3UPlaylistsProvider(backend.PlaylistsProvider):
    def __init__(self, backend, config):
        super().__init__(backend)

        ext_config = config[Extension.ext_name]
        if ext_config["playlists_dir"] is None:
            self._playlists_dir = Extension.get_data_dir(config)
        else:
            self._playlists_dir = path.expand_path(ext_config["playlists_dir"])
        if ext_config["base_dir"] is None:
            self._base_dir = self._playlists_dir
        else:
            self._base_dir = path.expand_path(ext_config["base_dir"])
        self._default_encoding = ext_config["default_encoding"]
        self._default_extension = ext_config["default_extension"]

    def as_list(self):
        result = []
        for entry in self._playlists_dir.iterdir():
            if entry.suffix not in [".m3u", ".m3u8"]:
                continue
            elif not entry.is_file():
                continue
            else:
                playlist_path = entry.relative_to(self._playlists_dir)
                result.append(translator.path_to_ref(playlist_path))
        result.sort(key=operator.attrgetter("name"))
        return result

    def create(self, name):
        path = translator.path_from_name(name.strip(), self._default_extension)
        try:
            with self._open(path, "w"):
                pass
            mtime = self._abspath(path).stat().st_mtime
        except OSError as e:
            log_environment_error(f"Error creating playlist {name!r}", e)
        else:
            return translator.playlist(path, [], mtime)

    def delete(self, uri):
        path = translator.uri_to_path(uri)
        if not self._is_in_basedir(path):
            logger.debug("Ignoring path outside playlist dir: %s", uri)
            return False
        try:
            self._abspath(path).unlink()
        except OSError as e:
            log_environment_error(f"Error deleting playlist {uri!r}", e)
            return False
        else:
            return True

    def get_items(self, uri):
        path = translator.uri_to_path(uri)
        if not self._is_in_basedir(path):
            logger.debug("Ignoring path outside playlist dir: %s", uri)
            return None
        try:
            with self._open(path, "r") as fp:
                items = translator.load_items(fp, self._base_dir)
        except OSError as e:
            log_environment_error(f"Error reading playlist {uri!r}", e)
        else:
            return items

    def lookup(self, uri):
        path = translator.uri_to_path(uri)
        if not self._is_in_basedir(path):
            logger.debug("Ignoring path outside playlist dir: %s", uri)
            return None
        try:
            with self._open(path, "r") as fp:
                items = translator.load_items(fp, self._base_dir)
            mtime = self._abspath(path).stat().st_mtime
        except OSError as e:
            log_environment_error(f"Error reading playlist {uri!r}", e)
        else:
            return translator.playlist(path, items, mtime)

    def refresh(self):
        pass  # nothing to do

    def save(self, playlist):
        path = translator.uri_to_path(playlist.uri)
        if not self._is_in_basedir(path):
            logger.debug("Ignoring path outside playlist dir: %s", playlist.uri)
            return None
        name = translator.name_from_path(path)
        try:
            with self._open(path, "w") as fp:
                translator.dump_items(playlist.tracks, fp)
            if playlist.name and playlist.name != name:
                orig_path = path
                path = translator.path_from_name(playlist.name.strip())
                path = path.with_suffix(orig_path.suffix)
                self._abspath(orig_path).rename(self._abspath(path))
            mtime = self._abspath(path).stat().st_mtime
        except OSError as e:
            log_environment_error(f"Error saving playlist {playlist.uri!r}", e)
        else:
            return translator.playlist(path, playlist.tracks, mtime)

    def _abspath(self, path):
        if not path.is_absolute():
            return self._playlists_dir / path
        else:
            return path

    def _is_in_basedir(self, local_path):
        local_path = self._abspath(local_path)
        return path.is_path_inside_base_dir(local_path, self._playlists_dir)

    def _open(self, path, mode="r"):
        if path.suffix == ".m3u8":
            encoding = "utf-8"
        else:
            encoding = self._default_encoding
        if not path.is_absolute():
            path = self._abspath(path)
        if not self._is_in_basedir(path):
            raise Exception(
                f"Path {path!r} is not inside playlist dir {self._playlist_dir!r}"
            )
        if "w" in mode:
            return replace(path, mode, encoding=encoding, errors="replace")
        else:
            return path.open(mode, encoding=encoding, errors="replace")



from mopidy import config


def test_core_schema_has_cache_dir():
    assert "cache_dir" in config._core_schema
    assert isinstance(config._core_schema["cache_dir"], config.Path)


def test_core_schema_has_config_dir():
    assert "config_dir" in config._core_schema
    assert isinstance(config._core_schema["config_dir"], config.Path)


def test_core_schema_has_data_dir():
    assert "data_dir" in config._core_schema
    assert isinstance(config._core_schema["data_dir"], config.Path)


def test_core_schema_has_max_tracklist_length():
    assert "max_tracklist_length" in config._core_schema
    max_tracklist_length_schema = config._core_schema["max_tracklist_length"]
    assert isinstance(max_tracklist_length_schema, config.Integer)
    assert max_tracklist_length_schema._minimum == 1



import os
import pathlib
from unittest import mock

import pytest

from mopidy.internal import xdg


@pytest.yield_fixture
def environ():
    patcher = mock.patch.dict(os.environ, clear=True)
    yield patcher.start()
    patcher.stop()


def test_cache_dir_default(environ):
    assert xdg.get_dirs()["XDG_CACHE_DIR"] == (
        pathlib.Path("~/.cache").expanduser()
    )


def test_cache_dir_from_env(environ):
    os.environ["XDG_CACHE_HOME"] = "/foo/bar"

    assert xdg.get_dirs()["XDG_CACHE_DIR"] == pathlib.Path("/foo/bar")


def test_config_dir_default(environ):
    assert xdg.get_dirs()["XDG_CONFIG_DIR"] == (
        pathlib.Path("~/.config").expanduser()
    )


def test_config_dir_from_env(environ):
    os.environ["XDG_CONFIG_HOME"] = "/foo/bar"

    assert xdg.get_dirs()["XDG_CONFIG_DIR"] == pathlib.Path("/foo/bar")


def test_data_dir_default(environ):
    assert xdg.get_dirs()["XDG_DATA_DIR"] == (
        pathlib.Path("~/.local/share").expanduser()
    )


def test_data_dir_from_env(environ):
    os.environ["XDG_DATA_HOME"] = "/foo/bar"

    assert xdg.get_dirs()["XDG_DATA_DIR"] == pathlib.Path("/foo/bar")


def test_user_dirs(environ, tmpdir):
    os.environ["XDG_CONFIG_HOME"] = str(tmpdir)

    with open(os.path.join(str(tmpdir), "user-dirs.dirs"), "wb") as fh:
        fh.write(b"# Some comments\n")
        fh.write(b'XDG_MUSIC_DIR="$HOME/Music2"\n')

    result = xdg.get_dirs()

    assert result["XDG_MUSIC_DIR"] == pathlib.Path("~/Music2").expanduser()
    assert "XDG_DOWNLOAD_DIR" not in result


def test_user_dirs_when_no_dirs_file(environ, tmpdir):
    os.environ["XDG_CONFIG_HOME"] = str(tmpdir)

    result = xdg.get_dirs()

    assert "XDG_MUSIC_DIR" not in result
    assert "XDG_DOWNLOAD_DIR" not in result



import contextlib
import re
import warnings

# Messages used in deprecation warnings are collected here so we can target
# them easily when ignoring warnings.
_MESSAGES = {
    # Deprecated features in core playback:
    "core.playback.play:tl_track_kwargs": (
        'playback.play() with "tl_track" argument is pending deprecation use '
        '"tlid" instead'
    ),
    # Deprecated features in core tracklist:
    "core.tracklist.add:tracks_arg": (
        'tracklist.add() "tracks" argument is deprecated'
    ),
    "core.tracklist.eot_track": (
        "tracklist.eot_track() is pending deprecation, use "
        "tracklist.get_eot_tlid()"
    ),
    "core.tracklist.next_track": (
        "tracklist.next_track() is pending deprecation, use "
        "tracklist.get_next_tlid()"
    ),
    "core.tracklist.previous_track": (
        "tracklist.previous_track() is pending deprecation, use "
        "tracklist.get_previous_tlid()"
    ),
}


def warn(msg_id, pending=False):
    if pending:
        category = PendingDeprecationWarning
    else:
        category = DeprecationWarning
    warnings.warn(_MESSAGES.get(msg_id, msg_id), category)


@contextlib.contextmanager
def ignore(ids=None):
    with warnings.catch_warnings():
        if isinstance(ids, str):
            ids = [ids]

        if ids:
            for msg_id in ids:
                msg = re.escape(_MESSAGES.get(msg_id, msg_id))
                warnings.filterwarnings("ignore", msg, DeprecationWarning)
        else:
            warnings.filterwarnings("ignore", category=DeprecationWarning)
        yield



import gzip
import json
import logging
import pathlib
import tempfile

from mopidy import models

logger = logging.getLogger(__name__)


def load(path):
    """
    Deserialize data from file.

    :param path: full path to import file
    :type path: pathlib.Path
    :return: deserialized data
    :rtype: dict
    """

    # TODO: raise an exception in case of error?
    if not path.is_file():
        logger.info("File does not exist: %s", path)
        return {}
    try:
        with gzip.open(str(path), "rb") as fp:
            return json.load(fp, object_hook=models.model_json_decoder)
    except (OSError, ValueError) as exc:
        logger.warning(f"Loading JSON failed: {exc}")
        return {}


def dump(path, data):
    """
    Serialize data to file.

    :param path: full path to export file
    :type path: pathlib.Path
    :param data: dictionary containing data to save
    :type data: dict
    """

    # TODO: cleanup directory/basename.* files.
    tmp = tempfile.NamedTemporaryFile(
        prefix=path.name + ".", dir=str(path.parent), delete=False
    )
    tmp_path = pathlib.Path(tmp.name)

    try:
        data_string = json.dumps(
            data, cls=models.ModelJSONEncoder, indent=2, separators=(",", ": ")
        )
        with gzip.GzipFile(fileobj=tmp, mode="wb") as fp:
            fp.write(data_string.encode())
        tmp_path.rename(path)
    finally:
        if tmp_path.exists():
            tmp_path.unlink()



import functools
import logging
import os
import urllib

import tornado.escape
import tornado.ioloop
import tornado.web
import tornado.websocket

import mopidy
from mopidy import core, models
from mopidy.internal import jsonrpc

logger = logging.getLogger(__name__)


def make_mopidy_app_factory(apps, statics):
    def mopidy_app_factory(config, core):
        if not config["http"]["csrf_protection"]:
            logger.warning(
                "HTTP Cross-Site Request Forgery protection is disabled"
            )
        allowed_origins = {
            x.lower() for x in config["http"]["allowed_origins"] if x
        }
        return [
            (
                r"/ws/?",
                WebSocketHandler,
                {
                    "core": core,
                    "allowed_origins": allowed_origins,
                    "csrf_protection": config["http"]["csrf_protection"],
                },
            ),
            (
                r"/rpc",
                JsonRpcHandler,
                {
                    "core": core,
                    "allowed_origins": allowed_origins,
                    "csrf_protection": config["http"]["csrf_protection"],
                },
            ),
            (
                r"/(.+)",
                StaticFileHandler,
                {"path": os.path.join(os.path.dirname(__file__), "data")},
            ),
            (r"/", ClientListHandler, {"apps": apps, "statics": statics}),
        ]

    return mopidy_app_factory


def make_jsonrpc_wrapper(core_actor):
    inspector = jsonrpc.JsonRpcInspector(
        objects={
            "core.get_uri_schemes": core.Core.get_uri_schemes,
            "core.get_version": core.Core.get_version,
            "core.history": core.HistoryController,
            "core.library": core.LibraryController,
            "core.mixer": core.MixerController,
            "core.playback": core.PlaybackController,
            "core.playlists": core.PlaylistsController,
            "core.tracklist": core.TracklistController,
        }
    )
    return jsonrpc.JsonRpcWrapper(
        objects={
            "core.describe": inspector.describe,
            "core.get_uri_schemes": core_actor.get_uri_schemes,
            "core.get_version": core_actor.get_version,
            "core.history": core_actor.history,
            "core.library": core_actor.library,
            "core.mixer": core_actor.mixer,
            "core.playback": core_actor.playback,
            "core.playlists": core_actor.playlists,
            "core.tracklist": core_actor.tracklist,
        },
        decoders=[models.model_json_decoder],
        encoders=[models.ModelJSONEncoder],
    )


def _send_broadcast(client, msg):
    # We could check for client.ws_connection, but we don't really
    # care why the broadcast failed, we just want the rest of them
    # to succeed, so catch everything.
    try:
        client.write_message(msg)
    except Exception as exc:
        logger.debug(
            f"Broadcast of WebSocket message to "
            f"{client.request.remote_ip} failed: {exc}"
        )
        # TODO: should this do the same cleanup as the on_message code?


class WebSocketHandler(tornado.websocket.WebSocketHandler):

    # XXX This set is shared by all WebSocketHandler objects. This isn't
    # optimal, but there's currently no use case for having more than one of
    # these anyway.
    clients = set()

    @classmethod
    def broadcast(cls, msg, io_loop):
        # This can be called from outside the Tornado ioloop, so we need to
        # safely cross the thread boundary by adding a callback to the loop.
        for client in cls.clients:
            # One callback per client to keep time we hold up the loop short
            io_loop.add_callback(
                functools.partial(_send_broadcast, client, msg)
            )

    def initialize(self, core, allowed_origins, csrf_protection):
        self.jsonrpc = make_jsonrpc_wrapper(core)
        self.allowed_origins = allowed_origins
        self.csrf_protection = csrf_protection

    def open(self):
        self.set_nodelay(True)
        self.clients.add(self)
        logger.debug("New WebSocket connection from %s", self.request.remote_ip)

    def on_close(self):
        self.clients.discard(self)
        logger.debug(
            "Closed WebSocket connection from %s", self.request.remote_ip
        )

    def on_message(self, message):
        if not message:
            return

        logger.debug(
            "Received WebSocket message from %s: %r",
            self.request.remote_ip,
            message,
        )

        try:
            response = self.jsonrpc.handle_json(
                tornado.escape.native_str(message)
            )
            if response and self.write_message(response):
                logger.debug(
                    "Sent WebSocket message to %s: %r",
                    self.request.remote_ip,
                    response,
                )
        except Exception as exc:
            logger.error(f"WebSocket request error: {exc}")
            self.close()

    def check_origin(self, origin):
        if not self.csrf_protection:
            return True
        return check_origin(origin, self.request.headers, self.allowed_origins)


def set_mopidy_headers(request_handler):
    request_handler.set_header("Cache-Control", "no-cache")
    request_handler.set_header("X-Mopidy-Version", mopidy.__version__.encode())


def check_origin(origin, request_headers, allowed_origins):
    if origin is None:
        logger.warning("HTTP request denied for missing Origin header")
        return False
    allowed_origins.add(request_headers.get("Host"))
    parsed_origin = urllib.parse.urlparse(origin).netloc.lower()
    # Some frameworks (e.g. Apache Cordova) use local files. Requests from
    # these files don't really have a sensible Origin so the browser sets the
    # header to something like 'file://' or 'null'. This results here in an
    # empty parsed_origin which we choose to allow.
    if parsed_origin and parsed_origin not in allowed_origins:
        logger.warning('HTTP request denied for Origin "%s"', origin)
        return False
    return True


class JsonRpcHandler(tornado.web.RequestHandler):
    def initialize(self, core, allowed_origins, csrf_protection):
        self.jsonrpc = make_jsonrpc_wrapper(core)
        self.allowed_origins = allowed_origins
        self.csrf_protection = csrf_protection

    def head(self):
        self.set_extra_headers()
        self.finish()

    def post(self):
        if self.csrf_protection:
            content_type = self.request.headers.get("Content-Type", "")
            if content_type != "application/json":
                self.set_status(415, "Content-Type must be application/json")
                return

        data = self.request.body
        if not data:
            return

        logger.debug(
            "Received RPC message from %s: %r", self.request.remote_ip, data
        )

        try:
            self.set_extra_headers()
            response = self.jsonrpc.handle_json(tornado.escape.native_str(data))
            if response and self.write(response):
                logger.debug(
                    "Sent RPC message to %s: %r",
                    self.request.remote_ip,
                    response,
                )
        except Exception as e:
            logger.error("HTTP JSON-RPC request error: %s", e)
            self.write_error(500)

    def set_extra_headers(self):
        set_mopidy_headers(self)
        self.set_header("Accept", "application/json")
        self.set_header("Content-Type", "application/json; utf-8")

    def options(self):
        if self.csrf_protection:
            origin = self.request.headers.get("Origin")
            if not check_origin(
                origin, self.request.headers, self.allowed_origins
            ):
                self.set_status(403, f"Access denied for origin {origin}")
                return

            self.set_header("Access-Control-Allow-Origin", f"{origin}")
            self.set_header("Access-Control-Allow-Headers", "Content-Type")

        self.set_status(204)
        self.finish()


class ClientListHandler(tornado.web.RequestHandler):
    def initialize(self, apps, statics):
        self.apps = apps
        self.statics = statics

    def get_template_path(self):
        return os.path.dirname(__file__)

    def get(self):
        set_mopidy_headers(self)

        names = set()
        for app in self.apps:
            names.add(app["name"])
        for static in self.statics:
            names.add(static["name"])
        names.discard("mopidy")

        self.render("data/clients.html", apps=sorted(list(names)))


class StaticFileHandler(tornado.web.StaticFileHandler):
    def set_extra_headers(self, path):
        set_mopidy_headers(self)


class AddSlashHandler(tornado.web.RequestHandler):
    @tornado.web.addslash
    def prepare(self):
        return super().prepare()



import configparser
import io
import xml.etree.ElementTree as elementtree  # noqa: N813

from mopidy.internal import validation


def parse(data):
    handlers = {
        detect_extm3u_header: parse_extm3u,
        detect_pls_header: parse_pls,
        detect_asx_header: parse_asx,
        detect_xspf_header: parse_xspf,
    }
    for detector, parser in handlers.items():
        if detector(data):
            return list(parser(data))
    return list(parse_urilist(data))  # Fallback


def detect_extm3u_header(data):
    return data[0:7].upper() == b"#EXTM3U"


def detect_pls_header(data):
    return data[0:10].lower() == b"[playlist]"


def detect_xspf_header(data):
    data = data[0:150]
    if b"xspf" not in data.lower():
        return False

    try:
        data = io.BytesIO(data)
        for _event, element in elementtree.iterparse(data, events=["start"]):
            return element.tag.lower() == "{http://xspf.org/ns/0/}playlist"
    except elementtree.ParseError:
        pass
    return False


def detect_asx_header(data):
    data = data[0:50]
    if b"asx" not in data.lower():
        return False

    try:
        data = io.BytesIO(data)
        for _event, element in elementtree.iterparse(data, events=["start"]):
            return element.tag.lower() == "asx"
    except elementtree.ParseError:
        pass
    return False


def parse_extm3u(data):
    # TODO: convert non URIs to file URIs.
    found_header = False
    for line in data.splitlines():
        if found_header or line.startswith(b"#EXTM3U"):
            found_header = True
        else:
            continue

        if not line.strip() or line.startswith(b"#"):
            continue

        try:
            line = line.decode()
        except UnicodeDecodeError:
            continue

        yield line.strip()


def parse_pls(data):
    # TODO: convert non URIs to file URIs.
    try:
        cp = configparser.RawConfigParser(strict=False)
        cp.read_string(data.decode())
    except configparser.Error:
        return

    for section in cp.sections():
        if section.lower() != "playlist":
            continue
        for i in range(cp.getint(section, "numberofentries")):
            yield cp.get(section, f"file{i + 1}").strip("\"'")


def parse_xspf(data):
    try:
        # Last element will be root.
        for _event, element in elementtree.iterparse(io.BytesIO(data)):
            element.tag = element.tag.lower()  # normalize
    except elementtree.ParseError:
        return

    ns = "http://xspf.org/ns/0/"
    path = f"{{{ns}}}tracklist/{{{ns}}}track"
    for track in element.iterfind(path):
        yield track.findtext(f"{{{ns}}}location")


def parse_asx(data):
    try:
        # Last element will be root.
        for _event, element in elementtree.iterparse(io.BytesIO(data)):
            element.tag = element.tag.lower()  # normalize
    except elementtree.ParseError:
        return

    for ref in element.findall("entry/ref[@href]"):
        yield ref.get("href", "").strip()

    for entry in element.findall("entry[@href]"):
        yield entry.get("href", "").strip()


def parse_urilist(data):
    for line in data.splitlines():
        if not line.strip() or line.startswith(b"#"):
            continue

        try:
            line = line.decode()
        except UnicodeDecodeError:
            continue

        try:
            validation.check_uri(line)
        except ValueError:
            continue

        yield line.strip()



import pytest

from mopidy.internal import playlists

BAD = b"foobarbaz"

EXTM3U = b"""#EXTM3U
#EXTINF:123, Sample artist - Sample title
file:///tmp/foo
#EXTINF:321,Example Artist - Example \xc5\xa7\xc5\x95
file:///tmp/bar

#EXTINF:213,Some Artist - Other title
file:///tmp/baz
"""

URILIST = b"""
file:///tmp/foo
# a comment \xc5\xa7\xc5\x95
file:///tmp/bar

file:///tmp/baz
"""

PLS = b"""[Playlist]
NumberOfEntries=3
File1="file:///tmp/foo"
Title1=Sample Title
Length1=123
Version=2

File2='file:///tmp/bar'
Title2=Example \xc5\xa7\xc5\x95
Length2=321
File3=file:///tmp/baz
Title3=Other title
Length3=213
Version=2
"""

ASX = b"""<ASX version="3.0">
  <TITLE>Example</TITLE>
  <ENTRY>
    <TITLE>Sample Title</TITLE>
    <REF href="file:///tmp/foo" />
  </ENTRY>
  <ENTRY>
    <TITLE>Example \xc5\xa7\xc5\x95</TITLE>
    <REF href="file:///tmp/bar" />
  </ENTRY>
  <ENTRY>
    <TITLE>Other title</TITLE>
    <REF href="file:///tmp/baz" />
  </ENTRY>
</ASX>
"""

SIMPLE_ASX = b"""<ASX version="3.0">
  <ENTRY href="file:///tmp/foo" />
  <ENTRY href="file:///tmp/bar" />
  <ENTRY href="file:///tmp/baz" />
</ASX>
"""

XSPF = b"""<?xml version="1.0" encoding="UTF-8"?>
<playlist version="1" xmlns="http://xspf.org/ns/0/">
  <trackList>
    <track>
      <title>Sample Title</title>
      <location>file:///tmp/foo</location>
    </track>
    <track>
      <title>Example \xc5\xa7\xc5\x95</title>
      <location>file:///tmp/bar</location>
    </track>
    <track>
      <title>Other title</title>
      <location>file:///tmp/baz</location>
    </track>
  </trackList>
</playlist>
"""

EXPECTED = ["file:///tmp/foo", "file:///tmp/bar", "file:///tmp/baz"]


@pytest.mark.parametrize(
    "detect_fn, data",
    [
        (playlists.detect_extm3u_header, EXTM3U),
        (playlists.detect_pls_header, PLS),
        (playlists.detect_asx_header, ASX),
        (playlists.detect_asx_header, SIMPLE_ASX),
        (playlists.detect_xspf_header, XSPF),
    ],
)
def test_detect_from_valid_header(detect_fn, data):
    assert detect_fn(data) is True


@pytest.mark.parametrize(
    "detect_fn",
    [
        playlists.detect_extm3u_header,
        playlists.detect_pls_header,
        playlists.detect_asx_header,
        playlists.detect_asx_header,
        playlists.detect_xspf_header,
    ],
)
def test_detect_from_invalid_header(detect_fn):
    assert detect_fn(BAD) is False


@pytest.mark.parametrize(
    "parse_fn, data",
    [
        (playlists.parse_extm3u, EXTM3U),
        (playlists.parse_pls, PLS),
        (playlists.parse_asx, ASX),
        (playlists.parse_asx, SIMPLE_ASX),
        (playlists.parse_xspf, XSPF),
        (playlists.parse_urilist, URILIST),
    ],
)
def test_parse_given_format_from_valid_data(parse_fn, data):
    assert list(parse_fn(data)) == EXPECTED


@pytest.mark.parametrize(
    "parse_fn",
    [
        playlists.parse_extm3u,
        playlists.parse_pls,
        playlists.parse_asx,
        playlists.parse_xspf,
        playlists.parse_urilist,
    ],
)
def test_parse_given_format_from_invalid_data(parse_fn):
    assert list(parse_fn(BAD)) == []


@pytest.mark.parametrize("data", [URILIST, EXTM3U, PLS, ASX, SIMPLE_ASX, XSPF])
def test_parse_any_format_from_valid_data(data):
    assert playlists.parse(data) == EXPECTED


def test_parse_from_invalid_data():
    assert playlists.parse(BAD) == []



import logging
import unittest
from unittest import mock

from mopidy.config import schemas, types

from tests import any_unicode


class ConfigSchemaTest(unittest.TestCase):
    def setUp(self):  # noqa: N802
        self.schema = schemas.ConfigSchema("test")
        self.schema["foo"] = mock.Mock()
        self.schema["bar"] = mock.Mock()
        self.schema["baz"] = mock.Mock()
        self.values = {"bar": "123", "foo": "456", "baz": "678"}

    def test_deserialize(self):
        self.schema.deserialize(self.values)

    def test_deserialize_with_missing_value(self):
        del self.values["foo"]

        result, errors = self.schema.deserialize(self.values)
        assert {"foo": any_unicode} == errors
        assert result.pop("foo") is None
        assert result.pop("bar") is not None
        assert result.pop("baz") is not None
        assert {} == result

    def test_deserialize_with_extra_value(self):
        self.values["extra"] = "123"

        result, errors = self.schema.deserialize(self.values)
        assert {"extra": any_unicode} == errors
        assert result.pop("foo") is not None
        assert result.pop("bar") is not None
        assert result.pop("baz") is not None
        assert {} == result

    def test_deserialize_with_deserialization_error(self):
        self.schema["foo"].deserialize.side_effect = ValueError("failure")

        result, errors = self.schema.deserialize(self.values)
        assert {"foo": "failure"} == errors
        assert result.pop("foo") is None
        assert result.pop("bar") is not None
        assert result.pop("baz") is not None
        assert {} == result

    def test_deserialize_with_multiple_deserialization_errors(self):
        self.schema["foo"].deserialize.side_effect = ValueError("failure")
        self.schema["bar"].deserialize.side_effect = ValueError("other")

        result, errors = self.schema.deserialize(self.values)
        assert {"foo": "failure", "bar": "other"} == errors
        assert result.pop("foo") is None
        assert result.pop("bar") is None
        assert result.pop("baz") is not None
        assert {} == result

    def test_deserialize_deserialization_unknown_and_missing_errors(self):
        self.values["extra"] = "123"
        self.schema["bar"].deserialize.side_effect = ValueError("failure")
        del self.values["baz"]

        result, errors = self.schema.deserialize(self.values)
        assert "unknown" in errors["extra"]
        assert "foo" not in errors
        assert "failure" in errors["bar"]
        assert "not found" in errors["baz"]

        assert "unknown" not in result
        assert "foo" in result
        assert result["bar"] is None
        assert result["baz"] is None

    def test_deserialize_deprecated_value(self):
        self.schema["foo"] = types.Deprecated()

        result, errors = self.schema.deserialize(self.values)
        assert ["bar", "baz"] == sorted(result.keys())
        assert "foo" not in errors


class MapConfigSchemaTest(unittest.TestCase):
    def test_conversion(self):
        schema = schemas.MapConfigSchema("test", types.LogLevel())
        result, errors = schema.deserialize({"foo.bar": "DEBUG", "baz": "INFO"})

        assert logging.DEBUG == result["foo.bar"]
        assert logging.INFO == result["baz"]


class DidYouMeanTest(unittest.TestCase):
    def test_suggestions(self):
        choices = ("enabled", "username", "password", "bitrate", "timeout")

        suggestion = schemas._did_you_mean("bitrate", choices)
        assert suggestion == "bitrate"

        suggestion = schemas._did_you_mean("bitrote", choices)
        assert suggestion == "bitrate"

        suggestion = schemas._did_you_mean("Bitrot", choices)
        assert suggestion == "bitrate"

        suggestion = schemas._did_you_mean("BTROT", choices)
        assert suggestion == "bitrate"

        suggestion = schemas._did_you_mean("btro", choices)
        assert suggestion is None



import logging
import os
import threading

import pykka

from mopidy import exceptions
from mopidy.audio import tags as tags_lib
from mopidy.audio import utils
from mopidy.audio.constants import PlaybackState
from mopidy.audio.listener import AudioListener
from mopidy.internal import process
from mopidy.internal.gi import GLib, GObject, Gst, GstPbutils

logger = logging.getLogger(__name__)

# This logger is only meant for debug logging of low level GStreamer info such
# as callbacks, event, messages and direct interaction with GStreamer such as
# set_state() on a pipeline.
gst_logger = logging.getLogger("mopidy.audio.gst")

_GST_PLAY_FLAGS_AUDIO = 0x02
_GST_PLAY_FLAGS_DOWNLOAD = 0x80

_GST_STATE_MAPPING = {
    Gst.State.PLAYING: PlaybackState.PLAYING,
    Gst.State.PAUSED: PlaybackState.PAUSED,
    Gst.State.NULL: PlaybackState.STOPPED,
}


# TODO: expose this as a property on audio?
class _Appsrc:

    """Helper class for dealing with appsrc based playback."""

    def __init__(self):
        self._signals = utils.Signals()
        self.reset()

    def reset(self):
        """Reset the helper.

        Should be called whenever the source changes and we are not setting up
        a new appsrc.
        """
        self.prepare(None, None, None, None)

    def prepare(self, caps, need_data, enough_data, seek_data):
        """Store info we will need when the appsrc element gets installed."""
        self._signals.clear()
        self._source = None
        self._caps = caps
        self._need_data_callback = need_data
        self._seek_data_callback = seek_data
        self._enough_data_callback = enough_data

    def configure(self, source):
        """Configure the supplied source for use.

        Should be called whenever we get a new appsrc.
        """
        source.set_property("caps", self._caps)
        source.set_property("format", "time")
        source.set_property("stream-type", "seekable")
        source.set_property("max-bytes", 1 << 20)  # 1MB
        source.set_property("min-percent", 50)

        if self._need_data_callback:
            self._signals.connect(
                source, "need-data", self._on_signal, self._need_data_callback
            )
        if self._seek_data_callback:
            self._signals.connect(
                source, "seek-data", self._on_signal, self._seek_data_callback
            )
        if self._enough_data_callback:
            self._signals.connect(
                source,
                "enough-data",
                self._on_signal,
                None,
                self._enough_data_callback,
            )

        self._source = source

    def push(self, buffer_):
        if self._source is None:
            return False

        if buffer_ is None:
            gst_logger.debug("Sending appsrc end-of-stream event.")
            result = self._source.emit("end-of-stream")
            return result == Gst.FlowReturn.OK
        else:
            result = self._source.emit("push-buffer", buffer_)
            return result == Gst.FlowReturn.OK

    def _on_signal(self, element, clocktime, func):
        # This shim is used to ensure we always return true, and also handles
        # that not all the callbacks have a time argument.
        if clocktime is None:
            func()
        else:
            func(utils.clocktime_to_millisecond(clocktime))
        return True


# TODO: expose this as a property on audio when #790 gets further along.
class _Outputs(Gst.Bin):
    def __init__(self):
        Gst.Bin.__init__(self)
        # TODO gst1: Set 'outputs' as the Bin name for easier debugging

        self._tee = Gst.ElementFactory.make("tee")
        self.add(self._tee)

        ghost_pad = Gst.GhostPad.new("sink", self._tee.get_static_pad("sink"))
        self.add_pad(ghost_pad)

    def add_output(self, description):
        # XXX This only works for pipelines not in use until #790 gets done.
        try:
            output = Gst.parse_bin_from_description(
                description, ghost_unlinked_pads=True
            )
        except GLib.GError as ex:
            logger.error(
                'Failed to create audio output "%s": %s', description, ex
            )
            raise exceptions.AudioException(bytes(ex))

        self._add(output)
        logger.info('Audio output set to "%s"', description)

    def _add(self, element):
        queue = Gst.ElementFactory.make("queue")
        self.add(element)
        self.add(queue)
        queue.link(element)
        self._tee.link(queue)


class SoftwareMixer:
    def __init__(self, mixer):
        self._mixer = mixer
        self._element = None
        self._last_volume = None
        self._last_mute = None
        self._signals = utils.Signals()

    def setup(self, element, mixer_ref):
        self._element = element
        self._mixer.setup(mixer_ref)

    def teardown(self):
        self._signals.clear()
        self._mixer.teardown()

    def get_volume(self):
        return int(round(self._element.get_property("volume") * 100))

    def set_volume(self, volume):
        self._element.set_property("volume", volume / 100.0)
        self._mixer.trigger_volume_changed(self.get_volume())

    def get_mute(self):
        return self._element.get_property("mute")

    def set_mute(self, mute):
        self._element.set_property("mute", bool(mute))
        self._mixer.trigger_mute_changed(self.get_mute())


class _Handler:
    def __init__(self, audio):
        self._audio = audio
        self._element = None
        self._pad = None
        self._message_handler_id = None
        self._event_handler_id = None

    def setup_message_handling(self, element):
        self._element = element
        bus = element.get_bus()
        bus.add_signal_watch()
        self._message_handler_id = bus.connect("message", self.on_message)

    def setup_event_handling(self, pad):
        self._pad = pad
        self._event_handler_id = pad.add_probe(
            Gst.PadProbeType.EVENT_BOTH, self.on_pad_event
        )

    def teardown_message_handling(self):
        bus = self._element.get_bus()
        bus.remove_signal_watch()
        bus.disconnect(self._message_handler_id)
        self._message_handler_id = None

    def teardown_event_handling(self):
        self._pad.remove_probe(self._event_handler_id)
        self._event_handler_id = None

    def on_message(self, bus, msg):
        if msg.type == Gst.MessageType.STATE_CHANGED:
            if msg.src != self._element:
                return
            old_state, new_state, pending_state = msg.parse_state_changed()
            self.on_playbin_state_changed(old_state, new_state, pending_state)
        elif msg.type == Gst.MessageType.BUFFERING:
            self.on_buffering(msg.parse_buffering(), msg.get_structure())
        elif msg.type == Gst.MessageType.EOS:
            self.on_end_of_stream()
        elif msg.type == Gst.MessageType.ERROR:
            error, debug = msg.parse_error()
            self.on_error(error, debug)
        elif msg.type == Gst.MessageType.WARNING:
            error, debug = msg.parse_warning()
            self.on_warning(error, debug)
        elif msg.type == Gst.MessageType.ASYNC_DONE:
            self.on_async_done()
        elif msg.type == Gst.MessageType.TAG:
            taglist = msg.parse_tag()
            self.on_tag(taglist)
        elif msg.type == Gst.MessageType.ELEMENT:
            if GstPbutils.is_missing_plugin_message(msg):
                self.on_missing_plugin(msg)
        elif msg.type == Gst.MessageType.STREAM_START:
            self.on_stream_start()

    def on_pad_event(self, pad, pad_probe_info):
        event = pad_probe_info.get_event()
        if event.type == Gst.EventType.SEGMENT:
            self.on_segment(event.parse_segment())
        return Gst.PadProbeReturn.OK

    def on_playbin_state_changed(self, old_state, new_state, pending_state):
        gst_logger.debug(
            "Got STATE_CHANGED bus message: old=%s new=%s pending=%s",
            old_state.value_name,
            new_state.value_name,
            pending_state.value_name,
        )

        if new_state == Gst.State.READY and pending_state == Gst.State.NULL:
            # XXX: We're not called on the last state change when going down to
            # NULL, so we rewrite the second to last call to get the expected
            # behavior.
            new_state = Gst.State.NULL
            pending_state = Gst.State.VOID_PENDING

        if pending_state != Gst.State.VOID_PENDING:
            return  # Ignore intermediate state changes

        if new_state == Gst.State.READY:
            return  # Ignore READY state as it's GStreamer specific

        new_state = _GST_STATE_MAPPING[new_state]
        old_state, self._audio.state = self._audio.state, new_state

        target_state = _GST_STATE_MAPPING.get(self._audio._target_state)
        if target_state is None:
            # XXX: Workaround for #1430, to be fixed properly by #1222.
            logger.warn("Race condition happened. See #1222 and #1430.")
            return
        if target_state == new_state:
            target_state = None

        logger.debug(
            "Audio event: state_changed(old_state=%s, new_state=%s, "
            "target_state=%s)",
            old_state,
            new_state,
            target_state,
        )
        AudioListener.send(
            "state_changed",
            old_state=old_state,
            new_state=new_state,
            target_state=target_state,
        )
        if new_state == PlaybackState.STOPPED:
            logger.debug("Audio event: stream_changed(uri=None)")
            AudioListener.send("stream_changed", uri=None)

        if "GST_DEBUG_DUMP_DOT_DIR" in os.environ:
            Gst.debug_bin_to_dot_file(
                self._audio._playbin, Gst.DebugGraphDetails.ALL, "mopidy"
            )

    def on_buffering(self, percent, structure=None):
        if self._audio._target_state < Gst.State.PAUSED:
            gst_logger.debug("Skip buffering during track change.")
            return

        if structure is not None and structure.has_field("buffering-mode"):
            buffering_mode = structure.get_enum(
                "buffering-mode", Gst.BufferingMode
            )
            if buffering_mode == Gst.BufferingMode.LIVE:
                return  # Live sources stall in paused.

        level = logging.getLevelName("TRACE")
        if percent < 10 and not self._audio._buffering:
            self._audio._playbin.set_state(Gst.State.PAUSED)
            self._audio._buffering = True
            level = logging.DEBUG
        if percent == 100:
            self._audio._buffering = False
            if self._audio._target_state == Gst.State.PLAYING:
                self._audio._playbin.set_state(Gst.State.PLAYING)
            level = logging.DEBUG

        gst_logger.log(
            level, "Got BUFFERING bus message: percent=%d%%", percent
        )

    def on_end_of_stream(self):
        gst_logger.debug("Got EOS (end of stream) bus message.")
        logger.debug("Audio event: reached_end_of_stream()")
        self._audio._tags = {}
        AudioListener.send("reached_end_of_stream")

    def on_error(self, error, debug):
        gst_logger.error(f"GStreamer error: {error.message}")
        gst_logger.debug(
            f"Got ERROR bus message: error={error!r} debug={debug!r}"
        )

        # TODO: is this needed?
        self._audio.stop_playback()

    def on_warning(self, error, debug):
        gst_logger.warning(f"GStreamer warning: {error.message}")
        gst_logger.debug(
            f"Got WARNING bus message: error={error!r} debug={debug!r}"
        )

    def on_async_done(self):
        gst_logger.debug("Got ASYNC_DONE bus message.")

    def on_tag(self, taglist):
        tags = tags_lib.convert_taglist(taglist)
        gst_logger.debug(
            f"Got TAG bus message: tags={tags_lib.repr_tags(tags)}"
        )

        # Postpone emitting tags until stream start.
        if self._audio._pending_tags is not None:
            self._audio._pending_tags.update(tags)
            return

        # TODO: Add proper tests for only emitting changed tags.
        unique = object()
        changed = []
        for key, value in tags.items():
            # Update any tags that changed, and store changed keys.
            if self._audio._tags.get(key, unique) != value:
                self._audio._tags[key] = value
                changed.append(key)

        if changed:
            logger.debug("Audio event: tags_changed(tags=%r)", changed)
            AudioListener.send("tags_changed", tags=changed)

    def on_missing_plugin(self, msg):
        desc = GstPbutils.missing_plugin_message_get_description(msg)
        debug = GstPbutils.missing_plugin_message_get_installer_detail(msg)
        gst_logger.debug("Got missing-plugin bus message: description=%r", desc)
        logger.warning("Could not find a %s to handle media.", desc)
        if GstPbutils.install_plugins_supported():
            logger.info(
                "You might be able to fix this by running: "
                'gst-installer "%s"',
                debug,
            )
        # TODO: store the missing plugins installer info in a file so we can
        # can provide a 'mopidy install-missing-plugins' if the system has the
        # required helper installed?

    def on_stream_start(self):
        gst_logger.debug("Got STREAM_START bus message")
        uri = self._audio._pending_uri
        logger.debug("Audio event: stream_changed(uri=%r)", uri)
        AudioListener.send("stream_changed", uri=uri)

        # Emit any postponed tags that we got after about-to-finish.
        tags, self._audio._pending_tags = self._audio._pending_tags, None
        self._audio._tags = tags or {}

        if tags:
            logger.debug("Audio event: tags_changed(tags=%r)", tags.keys())
            AudioListener.send("tags_changed", tags=tags.keys())

        if self._audio._pending_metadata:
            self._audio._playbin.send_event(self._audio._pending_metadata)
            self._audio._pending_metadata = None

    def on_segment(self, segment):
        gst_logger.debug(
            "Got SEGMENT pad event: "
            "rate=%(rate)s format=%(format)s start=%(start)s stop=%(stop)s "
            "position=%(position)s",
            {
                "rate": segment.rate,
                "format": Gst.Format.get_name(segment.format),
                "start": segment.start,
                "stop": segment.stop,
                "position": segment.position,
            },
        )
        position_ms = segment.position // Gst.MSECOND
        logger.debug("Audio event: position_changed(position=%r)", position_ms)
        AudioListener.send("position_changed", position=position_ms)


# TODO: create a player class which replaces the actors internals
class Audio(pykka.ThreadingActor):

    """
    Audio output through `GStreamer <https://gstreamer.freedesktop.org/>`_.
    """

    #: The GStreamer state mapped to :class:`mopidy.audio.PlaybackState`
    state = PlaybackState.STOPPED

    #: The software mixing interface :class:`mopidy.audio.actor.SoftwareMixer`
    mixer = None

    def __init__(self, config, mixer):
        super().__init__()

        self._config = config
        self._target_state = Gst.State.NULL
        self._buffering = False
        self._live_stream = False
        self._tags = {}
        self._pending_uri = None
        self._pending_tags = None
        self._pending_metadata = None

        self._playbin = None
        self._outputs = None
        self._queue = None
        self._about_to_finish_callback = None

        self._handler = _Handler(self)
        self._appsrc = _Appsrc()
        self._signals = utils.Signals()

        if mixer and self._config["audio"]["mixer"] == "software":
            self.mixer = pykka.traversable(SoftwareMixer(mixer))

    def on_start(self):
        self._thread = threading.current_thread()
        try:
            self._setup_preferences()
            self._setup_playbin()
            self._setup_outputs()
            self._setup_audio_sink()
        except GLib.GError as ex:
            logger.exception(ex)
            process.exit_process()

    def on_stop(self):
        self._teardown_mixer()
        self._teardown_playbin()

    def _setup_preferences(self):
        # TODO: move out of audio actor?
        # Fix for https://github.com/mopidy/mopidy/issues/604
        registry = Gst.Registry.get()
        jacksink = registry.find_feature("jackaudiosink", Gst.ElementFactory)
        if jacksink:
            jacksink.set_rank(Gst.Rank.SECONDARY)

    def _setup_playbin(self):
        playbin = Gst.ElementFactory.make("playbin")
        playbin.set_property("flags", _GST_PLAY_FLAGS_AUDIO)

        # TODO: turn into config values...
        playbin.set_property("buffer-size", 5 << 20)  # 5MB
        playbin.set_property("buffer-duration", 5 * Gst.SECOND)

        self._signals.connect(playbin, "source-setup", self._on_source_setup)
        self._signals.connect(
            playbin, "about-to-finish", self._on_about_to_finish
        )

        self._playbin = playbin
        self._handler.setup_message_handling(playbin)

    def _teardown_playbin(self):
        self._handler.teardown_message_handling()
        self._handler.teardown_event_handling()
        self._signals.disconnect(self._playbin, "about-to-finish")
        self._signals.disconnect(self._playbin, "source-setup")
        self._playbin.set_state(Gst.State.NULL)

    def _setup_outputs(self):
        # We don't want to use outputs for regular testing, so just install
        # an unsynced fakesink when someone asks for a 'testoutput'.
        if self._config["audio"]["output"] == "testoutput":
            self._outputs = Gst.ElementFactory.make("fakesink")
        else:
            self._outputs = _Outputs()
            try:
                self._outputs.add_output(self._config["audio"]["output"])
            except exceptions.AudioException:
                process.exit_process()  # TODO: move this up the chain

        self._handler.setup_event_handling(self._outputs.get_static_pad("sink"))

    def _setup_audio_sink(self):
        audio_sink = Gst.ElementFactory.make("bin", "audio-sink")
        queue = Gst.ElementFactory.make("queue")
        volume = Gst.ElementFactory.make("volume")

        # Queue element to buy us time between the about-to-finish event and
        # the actual switch, i.e. about to switch can block for longer thanks
        # to this queue.

        # TODO: See if settings should be set to minimize latency. Previous
        # setting breaks appsrc, and settings before that broke on a few
        # systems. So leave the default to play it safe.
        buffer_time = self._config["audio"]["buffer_time"]
        if buffer_time is not None and buffer_time > 0:
            queue.set_property("max-size-time", buffer_time * Gst.MSECOND)

        audio_sink.add(queue)
        audio_sink.add(self._outputs)
        audio_sink.add(volume)

        queue.link(volume)
        volume.link(self._outputs)

        if self.mixer:
            self.mixer.setup(volume, self.actor_ref.proxy().mixer)

        ghost_pad = Gst.GhostPad.new("sink", queue.get_static_pad("sink"))
        audio_sink.add_pad(ghost_pad)

        self._playbin.set_property("audio-sink", audio_sink)
        self._queue = queue

    def _teardown_mixer(self):
        if self.mixer:
            self.mixer.teardown()

    def _on_about_to_finish(self, element):
        if self._thread == threading.current_thread():
            logger.error(
                "about-to-finish in actor, aborting to avoid deadlock."
            )
            return

        gst_logger.debug("Got about-to-finish event.")
        if self._about_to_finish_callback:
            logger.debug("Running about-to-finish callback.")
            self._about_to_finish_callback()

    def _on_source_setup(self, element, source):
        gst_logger.debug(
            "Got source-setup signal: element=%s", source.__class__.__name__
        )

        if source.get_factory().get_name() == "appsrc":
            self._appsrc.configure(source)
        else:
            self._appsrc.reset()

        if self._live_stream and hasattr(source.props, "is_live"):
            gst_logger.debug("Enabling live stream mode")
            source.set_live(True)

        utils.setup_proxy(source, self._config["proxy"])

    def set_uri(self, uri, live_stream=False, download=False):
        """
        Set URI of audio to be played.

        You *MUST* call :meth:`prepare_change` before calling this method.

        :param uri: the URI to play
        :type uri: string
        :param live_stream: disables buffering, reducing latency for stream,
            and discarding data when paused
        :type live_stream: bool
        :param download: enables "download" buffering mode
        :type download: bool
        """

        # XXX: Hack to workaround issue on Mac OS X where volume level
        # does not persist between track changes. mopidy/mopidy#886
        if self.mixer is not None:
            current_volume = self.mixer.get_volume()
        else:
            current_volume = None

        flags = _GST_PLAY_FLAGS_AUDIO
        if download:
            flags |= _GST_PLAY_FLAGS_DOWNLOAD

        logger.debug(f"Flags: {flags}")
        if live_stream and download:
            logger.warning(
                "Ambiguous buffering flags: "
                "'live_stream' and 'download' should not both be set."
            )

        self._pending_uri = uri
        self._pending_tags = {}
        self._live_stream = live_stream
        self._playbin.set_property("flags", flags)
        self._playbin.set_property("uri", uri)

        if self.mixer is not None and current_volume is not None:
            self.mixer.set_volume(current_volume)

    def set_appsrc(
        self, caps, need_data=None, enough_data=None, seek_data=None
    ):
        """
        Switch to using appsrc for getting audio to be played.

        You *MUST* call :meth:`prepare_change` before calling this method.

        :param caps: GStreamer caps string describing the audio format to
            expect
        :type caps: string
        :param need_data: callback for when appsrc needs data
        :type need_data: callable which takes data length hint in ms
        :param enough_data: callback for when appsrc has enough data
        :type enough_data: callable
        :param seek_data: callback for when data from a new position is needed
            to continue playback
        :type seek_data: callable which takes time position in ms
        """
        self._appsrc.prepare(
            Gst.Caps.from_string(caps), need_data, enough_data, seek_data
        )
        uri = "appsrc://"
        self._pending_uri = uri
        self._playbin.set_property("uri", uri)

    def emit_data(self, buffer_):
        """
        Call this to deliver raw audio data to be played.

        If the buffer is :class:`None`, the end-of-stream token is put on the
        playbin. We will get a GStreamer message when the stream playback
        reaches the token, and can then do any end-of-stream related tasks.

        Note that the URI must be set to ``appsrc://`` for this to work.

        Returns :class:`True` if data was delivered.

        :param buffer_: buffer to pass to appsrc
        :type buffer_: :class:`Gst.Buffer` or :class:`None`
        :rtype: boolean
        """
        return self._appsrc.push(buffer_)

    def set_about_to_finish_callback(self, callback):
        """
        Configure audio to use an about-to-finish callback.

        This should be used to achieve gapless playback. For this to work the
        callback *MUST* call :meth:`set_uri` with the new URI to play and
        block until this call has been made. :meth:`prepare_change` is not
        needed before :meth:`set_uri` in this one special case.

        :param callable callback: Callback to run when we need the next URI.
        """
        self._about_to_finish_callback = callback

    def get_position(self):
        """
        Get position in milliseconds.

        :rtype: int
        """
        success, position = self._playbin.query_position(Gst.Format.TIME)

        if not success:
            # TODO: take state into account for this and possibly also return
            # None as the unknown value instead of zero?
            logger.debug("Position query failed")
            return 0

        return utils.clocktime_to_millisecond(position)

    def set_position(self, position):
        """
        Set position in milliseconds.

        :param position: the position in milliseconds
        :type position: int
        :rtype: :class:`True` if successful, else :class:`False`
        """
        # TODO: double check seek flags in use.
        gst_position = utils.millisecond_to_clocktime(position)
        gst_logger.debug("Sending flushing seek: position=%r", gst_position)
        # Send seek event to the queue not the playbin. The default behavior
        # for bins is to forward this event to all sinks. Which results in
        # duplicate seek events making it to appsrc. Since elements are not
        # allowed to act on the seek event, only modify it, this should be safe
        # to do.
        result = self._queue.seek_simple(
            Gst.Format.TIME, Gst.SeekFlags.FLUSH, gst_position
        )
        return result

    def start_playback(self):
        """
        Notify GStreamer that it should start playback.

        :rtype: :class:`True` if successfull, else :class:`False`
        """
        return self._set_state(Gst.State.PLAYING)

    def pause_playback(self):
        """
        Notify GStreamer that it should pause playback.

        :rtype: :class:`True` if successfull, else :class:`False`
        """
        return self._set_state(Gst.State.PAUSED)

    def prepare_change(self):
        """
        Notify GStreamer that we are about to change state of playback.

        This function *MUST* be called before changing URIs or doing
        changes like updating data that is being pushed. The reason for this
        is that GStreamer will reset all its state when it changes to
        :attr:`Gst.State.READY`.
        """
        return self._set_state(Gst.State.READY)

    def stop_playback(self):
        """
        Notify GStreamer that is should stop playback.

        :rtype: :class:`True` if successfull, else :class:`False`
        """
        return self._set_state(Gst.State.NULL)

    def wait_for_state_change(self):
        """Block until any pending state changes are complete.

        Should only be used by tests.
        """
        self._playbin.get_state(timeout=Gst.CLOCK_TIME_NONE)

    def enable_sync_handler(self):
        """Enable manual processing of messages from bus.

        Should only be used by tests.
        """

        def sync_handler(bus, message):
            self._handler.on_message(bus, message)
            return Gst.BusSyncReply.DROP

        bus = self._playbin.get_bus()
        bus.set_sync_handler(sync_handler)

    def _set_state(self, state):
        """
        Internal method for setting the raw GStreamer state.

        .. digraph:: gst_state_transitions

            graph [rankdir="LR"];
            node [fontsize=10];

            "NULL" -> "READY"
            "PAUSED" -> "PLAYING"
            "PAUSED" -> "READY"
            "PLAYING" -> "PAUSED"
            "READY" -> "NULL"
            "READY" -> "PAUSED"

        :param state: State to set playbin to. One of: `Gst.State.NULL`,
            `Gst.State.READY`, `Gst.State.PAUSED` and `Gst.State.PLAYING`.
        :type state: :class:`Gst.State`
        :rtype: :class:`True` if successfull, else :class:`False`
        """
        if state < Gst.State.PAUSED:
            self._buffering = False

        self._target_state = state
        result = self._playbin.set_state(state)
        gst_logger.debug(
            "Changing state to %s: result=%s",
            state.value_name,
            result.value_name,
        )

        if result == Gst.StateChangeReturn.FAILURE:
            logger.warning(
                "Setting GStreamer state to %s failed", state.value_name
            )
            return False
        # TODO: at this point we could already emit stopped event instead
        # of faking it in the message handling when result=OK
        return True

    # TODO: bake this into setup appsrc perhaps?
    def set_metadata(self, track):
        """
        Set track metadata for currently playing song.

        Only needs to be called by sources such as ``appsrc`` which do not
        already inject tags in playbin, e.g. when using :meth:`emit_data` to
        deliver raw audio data to GStreamer.

        :param track: the current track
        :type track: :class:`mopidy.models.Track`
        """
        taglist = Gst.TagList.new_empty()
        artists = [a for a in (track.artists or []) if a.name]

        def set_value(tag, value):
            gobject_value = GObject.Value()
            gobject_value.init(GObject.TYPE_STRING)
            gobject_value.set_string(value)
            taglist.add_value(Gst.TagMergeMode.REPLACE, tag, gobject_value)

        # Default to blank data to trick shoutcast into clearing any previous
        # values it might have.
        # TODO: Verify if this works at all, likely it doesn't.
        set_value(Gst.TAG_ARTIST, " ")
        set_value(Gst.TAG_TITLE, " ")
        set_value(Gst.TAG_ALBUM, " ")

        if artists:
            set_value(Gst.TAG_ARTIST, ", ".join(a.name for a in artists))

        if track.name:
            set_value(Gst.TAG_TITLE, track.name)

        if track.album and track.album.name:
            set_value(Gst.TAG_ALBUM, track.album.name)

        gst_logger.debug(
            "Sending TAG event for track %r: %r", track.uri, taglist.to_string()
        )
        event = Gst.Event.new_tag(taglist)
        if self._pending_uri:
            self._pending_metadata = event
        else:
            self._playbin.send_event(event)

    def get_current_tags(self):
        """
        Get the currently playing media's tags.

        If no tags have been found, or nothing is playing this returns an empty
        dictionary. For each set of tags we collect a tags_changed event is
        emitted with the keys of the changes tags. After such calls users may
        call this function to get the updated values.

        :rtype: {key: [values]} dict for the current media.
        """
        # TODO: should this be a (deep) copy? most likely yes
        # TODO: should we return None when stopped?
        # TODO: support only fetching keys we care about?
        return self._tags
